import requests
import os
import json
import logging
import tempfile
import uuid
from requests.auth import HTTPBasicAuth

def fetchBridgeToken(link,branch,username,token):
        list=link.split("/")
        api = f'https://api.github.com/repos/{list[3]}/{list[4]}/contents/{"/".join(list[7:])}?ref={branch}'
        try:
            response = requests.get(api,auth = HTTPBasicAuth(username, token))
            json_response=json.loads(response.text)
            content_api=json_response["download_url"]
            response_content=requests.get(content_api)
            log_dir = str(os.path.join(tempfile.gettempdir(), 'XML'))
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
            suiteLogsLoc = os.path.join(log_dir, 'XML_' + str(uuid.uuid4()) + '.xml')
            print(response_content.text)
            with open(suiteLogsLoc, "w+") as f:
                f.write(response_content.text)
            return suiteLogsLoc
        except Exception as e:                  
            logging.info("Some Error while running the API")

if __name__ == "__main__":
    fetchBridgeToken("GIT:https://github.com/Gemini-Solutions/gempyp/blob/dev/tests/configTest/Gempyp_Test_suite.xml:dev:username:token")


import json
import requests
import logging
import os
import sys


upload_file_api = "https://apis-beta.gemecosystem.com/v1/upload/file"
upload_data_api = "https://apis-beta.gemecosystem.com/v1/upload/data"
delete_file = "https://apis-beta.gemecosystem.com/v1/file/tag"

def upload_to_s3(api=None, tag=None, file=None, data=None, s_run_id=None, folder=None, username=None, bearer_token=None, bridge_token=None):
    logging.info("In Upload section")
    if bearer_token is not None:
        headers = {"Authorization": "Bearer {}".format(bearer_token)}
    elif bridge_token is not None:
        if username is None:
            logging.ERROR("------ Please pass username along with Bridge token -----")
            sys.exit()
        headers = {"bridgeToken": bridge_token, "username": username}
    params = dict()
    if folder is not None:
        params["folder"] = folder
    if tag is not None: 
        params["tag"] = tag
    if s_run_id is not None:
        params["s_run_id"] = s_run_id
    if data is not None:
        logging.info("------ Uploading data to s3 -----")
        if api is None:
            api = "https://apis-beta.gemecosystem.com/v1/upload/data"
        headers["Content-Type"] = "text/plain"
        params["file"] = file
        response = requests.post(api, params=params, data=data, headers=headers)
        data = json.loads(response.text)    
        if response.status_code == 200:
            data = json.loads(response.text)    
            data_info = data["data"]
            print(data_info)
            return data_info
    elif file is not None:
        file = file.split(",")
        files = list()
        logging.info("---- Uploading file to s3 ----")
        if api is None:
            api = "https://apis-beta.gemecosystem.com/v1/upload/file"
        for f in file:
            if not os.path.isfile(f):
                logging.error("Path of file invalid - ", f)
                continue
            files.append(("file", open(f, "rb")))
        response = requests.post(api, files=files, params=params, headers=headers)  
        data = json.loads(response.text)    
        if response.status_code == 200:
            data = json.loads(response.text)    
            file_info = data["data"]
            return file_info

def download_from_s3(api, bearer_token=None, bridge_token=None, username=None, id=None):
    logging.info("In Download section")
    if bearer_token is not None:
        headers = {"Authorization": "Bearer {}".format(bearer_token)}
    elif bridge_token is not None:
        if username is None:
            logging.ERROR("------ Please pass username along with Bridge token -----")
            sys.exit()
        headers = {"bridgeToken": bridge_token, "username": username}
    params = dict()
    if '?' not in api and id is None:
        logging.error("Please pass download url or the id")
        sys.exit()
    elif '?' in api:
        api_query = api.split('?')
        params = api_query[1][3:]
        api = api_query[0]
        params = {"id": "{}".format(params)}
    elif '?' not in api and id is not None:
        params = {"id": id}
    response = requests.get(api, params=params, headers=headers)  
    return response.text


if __name__ == "__main__":
    print(download_from_s3(bridge_token="9998284c-b61e-45d1-accd-2ce7d6f8f0a81670592834097"))


import traceback
import requests
import logging
from gempyp.config import DefaultSettings
import logging
import re
import json

not_uploaded = []
suite_data = []
flag = False
suite_uploaded = False
list_of_testcase = []
respon = {}

stat = {"TOTAL": 0, "PASS": 0, "FAIL": 0, "EXE": 0, "ERR": 0, "WARN": 0}
def _getHeaders(bridge_token, user_name):
    """
    for getting the bridgeToken in _sendData method
    """
    return {"Content-Type": "application/json", "bridgeToken": bridge_token, "username": user_name}

def checkingData(run_id, bridge_token, user_name):
    try:
        url = DefaultSettings.getUrls("suite-exe-api") + "?s_run_id=" + run_id
        response = _sendData("", url , bridge_token, user_name, "GET")
        if response.status_code == 200:
            logging.info("Retrying to update and add testcases data present in DB")
            logging.info("Get Request successfull")
            global respon, suite_uploaded, list_of_testcase
            suite_uploaded = True
            respon = response.json()
            list_of_testcase = [s[:-37] for s in respon['data']['testcase_details']]
            return response._content
        else:
            return "failed"
    except Exception as e:
        traceback.print_exc()
        return "failed"

def sendSuiteData(payload, bridge_token, user_name, mode="POST"):
    """
    for checking the sendSuiteData api response
    """
    try:
        if len(respon) != 0:
            payload = dataAlter(payload)
        payload = noneRemover(payload)
        response = _sendData(payload, DefaultSettings.getUrls("suite-exe-api"), bridge_token, user_name, mode)

        if response and response.status_code == 201:
            global suite_uploaded
            logging.info("Suite data uploaded successfully")
            suite_uploaded = True
            if payload in suite_data:
                suite_data.remove(payload)
        elif response and response.status_code == 200:
            logging.info("Suite Data updated Successfully")
        else:
            logging.info("Suite data is not uploaded")
            if payload not in suite_data:
                suite_data.append(payload)
                
    except Exception as e:
        logging.error(traceback.format_exc())

def sendTestcaseData(payload, bridge_token, user_name):
    """
    for checking the sendTestCaseData api response
    """

    try:
        method = "POST"
        payload = json.loads(payload)
        tc_run_id = payload["tc_run_id"]
        if tc_run_id[:-37] in list_of_testcase:
        #checking whether testcase present in previous run 
            payload, method = getTestcase(payload, method, bridge_token, user_name)
        payload = json.dumps(payload)
        response = _sendData(payload, DefaultSettings.getUrls("test-exe-api"), bridge_token, user_name, method)
        ### Applying regex to the response

        x = re.search("already present",response.text,re.IGNORECASE)
        if response.status_code == 201:
            logging.info("data uploaded successfully")
            if payload in not_uploaded:
                not_uploaded.remove(payload)
        elif response.status_code == 200:
            logging.info("data updated successfully")
            if payload in not_uploaded:
                not_uploaded.remove(payload)
    ### code for adding unuploaded testcases
        else:
            if payload not in not_uploaded:
                not_uploaded.append(payload)
                logging.info("Testcase data is not uploaded")
                if x != None:
                    global flag
                    flag = True

    except Exception as e:
        logging.error(traceback.format_exc())


def _sendData(payload, url, bridge_token, user_name, method="POST"):
    
    """
    calling the api to upload the data into database
    takes data we need to send(payload),bridgeToken,userName and method as argument
    """
    # Not needed anymore as we will be reuploading the data to db.
    # if DefaultSettings.count > 3:         
    #     logging.warning("Incorrect bridgetoken/username or APIs are down. Skipping Data upload.")
    #     return None
    response = requests.request(
        method=method,
        url=url,
        data=payload,
        headers=_getHeaders(bridge_token, user_name),
    )
    logging.info(f"status: {response.status_code}")
    return response

def noneRemover(payload):
    
    data = json.loads(payload)
    for key,value in dict(data).items():
        if value == "-":
            del data[key]
    return json.dumps(data)

# code for changing suite data
def dataAlter(payload):
    payload = json.loads(payload)
    global respon, stat
    # changing start time
    payload['s_start_time'] = respon['data']['s_start_time']
    # adding the testcase analytics of both run
    payload['testcase_info'] = {i: payload['testcase_info'].get(i, 0) + respon['data']['testcase_info'].get(i, 0)
    for i in set(payload['testcase_info']).union(respon['data']['testcase_info'])}
    # updating the testcase analytics according to new run
    for key, value in stat.items():
        if key in payload['testcase_info']:
            payload['testcase_info'][key] += value
    # making testcase info in order
    prio_list = ['TOTAL', 'PASS', 'FAIL'] 
    sorted_dict = {}
    for key in prio_list:
        if key in payload['testcase_info'] :
            sorted_dict[key] = payload['testcase_info'][key]
            if key == "PASS" or key == "FAIL":
                pass
            else:
                if sorted_dict[key] == 0:
                    sorted_dict.pop(key)
            payload['testcase_info'].pop(key)
        elif key == "PASS" or key == 'FAIL':
                sorted_dict[key] = 0
    sorted_dict.update(payload['testcase_info']) 
    status = "PASS"
    if sorted_dict["PASS"] != sorted_dict["TOTAL"]:
        status = "FAIL"
    payload['testcase_info'] = sorted_dict
    payload['status'] = status
    # updating the expected testcase
    payload['expected_testcases'] += respon['data']['expected_testcases']
    # updating the suite status of according to new run
    if payload['status'] != "PASS":
        pass
    else:
        payload['status'] = respon['data']['status']
    return  json.dumps(payload)

# code for checking and updating testcase details
def getTestcase(payload, method, bridge_token, user_name):
    global flag
    for i in respon['data']['testcase_details']:
        if payload['tc_run_id'][:-37] in i:
            url = DefaultSettings.getUrls("test-exe-api") + "?tc_run_id=" + i
            response = _sendData(" ",url, bridge_token, user_name, "GET")
            if response.status_code == 200:
                logging.info("Testcase get request successfull")
                global flag
                flag = True
                if payload in not_uploaded:
                    not_uploaded.remove(payload)
            res = response.json()
            status = res['data']['status']
            stat[status] -= 1
            payload['tc_run_id'] = i
            stat["TOTAL"] -=1
            method = "PUT"
    return payload, method




import json
import logging as logger
import ast
import re
from urllib import response
import gempyp.pyprest.compareFunctions as cf



def formatRespBody(response_body):
    """Encode/decode the response body"""
    if not isinstance(response_body, list):
        try:
            formatted_response_body = json.loads(response_body.decode("utf-8"))
            logger.info("keycheck ----- 1")
        except Exception:
            try:
                formatted_response_body = json.loads(ast.literal_eval(response_body.decode("utf-8")))
                logger.info("keycheck ----- 2")
            except Exception:
                try: 
                    formatted_response_body = json.loads(str(response_body.encode("utf-8")))
                    if "b'" in formatted_response_body:
                        formatted_response_body = formatted_response_body.strip("b'").strip("'")
                    logger.info("keycheck ----- 3")
                except Exception:
                    try:
                        formatted_response_body = json.loads(response_body)
                    except Exception as e:
                        formatted_response_body = response_body

    else:
        formatted_response_body = response_body
    return formatted_response_body


# not able to implement recursion in case of other 
def fetchValueOfKey(json_, key_partition_list, key_search_result, final_key_value={}):
    """
    Get values of the required keys from the response json"""

    logger.info("===========Fetching values from response ============")
    regex_each = re.compile(r".*\[\beach\b\]")
    regex_int = re.compile(r".*\[\d+\]")
    actual_key = ".".join(key_partition_list)
    flag = 0
    if key_search_result == "FOUND" or "MISSING IN " in key_search_result:
        for key in key_partition_list:
            if re.match(regex_each, key):
                each_value_list = []
                br_index = key.find('[')
                key_val = key[:br_index]
                if key_val != "response" or key_val != "legacy":
                    json_ = json_[key_val]
                for each_value in json_:
                    index_of_key = key_partition_list.index(key)
                    value_returned = getValuesForEach(each_value, key_partition_list[int(index_of_key) + 1:])
                    each_value_list.append(value_returned)
                del key_partition_list[int(index_of_key) + 1:]
                flag = 1
            elif re.match(regex_int, key):
                br_start = key.find('[')
                br_end = key.find(']')
                key_val = key[:br_start]
                key_num = int(key[br_start + 1:br_end])
                key_num, json_ = getNestedListData(key, json_, key_val)

                if key_val == "response" or key_val=="legacy" and isinstance(json_, list) :
                    json_ = json_[key_num]
                else:
                    json_ = json_[key_val][key_num]
            else:
                if isinstance(json_, str) and json_ != "":
                    json_ = json.dumps(json_)
                if json_ != "":
                    if "response" == key.lower() or "legacy" == key.lower():
                        json_ = json_
                    else:
                        json_ = json_[key]
                if json_ is None:
                    json_ = "null"
        if flag != 1:
            final_key_value[actual_key] = json_
        else:
            final_key_value[actual_key] = each_value_list
        
        return final_key_value


def getNestedListData(i, json_data, key_val):
    """parse nested lists in response"""
    # check if response is empty or not, if response is empty, how did it reach here?

    br_start = i.find('[')
    br_end = i.find(']')
    key_num = int(i[br_start + 1:br_end])
    if key_val.lower() == 'legacy' or key_val.lower() == 'response':
        if isinstance(json_data[key_num], list):
            i = i[br_end + 1:]
            json_data = json_data[key_num]
            return getNestedListData(i, json_data, key_val)
        else:
            return key_num, json_data
    else:
        if isinstance(json_data[key_val][key_num], list):
            i = i[br_end + 1:]
            json_data = json_data[key_val][key_num]
            return getNestedListData(i, json_data, key_val)
        else:
            return key_num, json_data


def dispatch_dict():
    dispatch = {
        'to': cf.compareTo,
        'notto': cf.compareNotTo,
        'not to': cf.compareNotTo,
        'not_to': cf.compareNotTo,
        'in': cf.compareIn,
        'notin': cf.compareNotIn,
        'not in': cf.compareNotIn,
        'not_in': cf.compareNotIn,
        'contains': cf.compareContains,
        'not_supported': cf.noOperator,
    }
    return dispatch


def compare(reporter_obj, key, operator, value, key_val_dict, tolerance=0.1, isLegacyPresent = False, isLegacyResponse = False):

    # operators ----- to, notto, not_to, not to, contains, in, except
    gp = reporter_obj

    dispatch = dispatch_dict()
    if operator in list(dispatch.keys()):
        return dispatch[operator](gp, key, value, key_val_dict, tolerance, isLegacyPresent, isLegacyResponse)
    else:
        return dispatch["not_supported"](gp)

class getKeyList:
    def __init__(self):
        self.keyList = []

    def parseList(self,response):
        if isinstance(response,list):
            for i in range(len(response)):
                if isinstance(response[i],dict):
                    self.parseDict(response[i])
                elif isinstance(response[i],list):
                    self.parseList(response[i])


    def parseDict(self,response):
        if isinstance(response,dict):
            for k,v in response.items():
                self.keyList.append(k)
                if isinstance(v,dict):
                    self.parseDict(v)
                elif isinstance(v,list):
                    self.parseList(v)
        elif isinstance(response,list):
            self.parseList(response)


    def getKeys(self, response_body):
        self.parseDict(response_body)
        return list(set(self.keyList)) 
    


import mysql.connector
import os
from configparser import ConfigParser
from typing import Dict
import pandas as pd
from gempyp.engine.baseTemplate import TestcaseReporter as Base
from gempyp.libs.enums.status import status
from gempyp.libs.common import readPath
from gempyp.dv.dvReporting import writeToReport
from telnetlib import STATUS
import traceback
import pandas as pd
import logging
import math
import numpy
import pg8000

class DvRunner(Base):

    def __init__(self, data):
        
        self.data = data

        self.configData: Dict = self.data.get("config_data")
        self.logger = data["config_data"]["LOGGER"] if "LOGGER" in data["config_data"].keys() else logging

        self.logger.info("---------------------Inside DVM FRAMEWORK------------------------")

        self.logger.info(f"-------Executing testcase - \"{self.data['config_data']['NAME']}\"---------")
        # # set vars
        self.setVars()
        
        # # setting self.reporter object
        self.logger.info("--------------------Report object created ------------------------")
        self.reporter = Base(project_name=self.project, testcase_name=self.tcname)
        
    def dvEngine(self):
        
        try:
            column =[]
            try:
                if self.configData["DATABASE"].lower() == 'custom': 
                    pass
                else:
                    configPath= self.configData["DB_CONFIG_PATH"]
                    configur = ConfigParser()
                    config = readPath(configPath)
                    configur.read(config)
            except Exception as e:
                self.reporter.addRow("Config File","Path is not Correct",status.FAIL)
                traceback.print_exc()
                output = writeToReport(self)
                return output, None
            try:
                if self.configData["DATABASE"].lower() == 'custom': 
                    conn_string = self.configData['SOURCE_CONN']
                    ind = conn_string.index('host')
                    count = 0
                    li = []
                    for i in range(ind,len(conn_string)):
                        if conn_string[i] == "'":
                            count +=1
                            li.append(i)
                            if count == 2:
                                break
                    host = conn_string[li[0]+1:li[1]]
                    userCred = {'host': host}
                    conn_string2 = self.configData['TARGET_CONN']
                    ind1 = conn_string2.index('host')
                    count = 0
                    li = []
                    for i in range(ind1,len(conn_string2)):
                        if conn_string2[i] == "'":
                            count +=1
                            li.append(i)
                            if count == 2:
                                break
                    host = conn_string2[li[0]+1:li[1]]
                    targetCred = {'host': host}
                    
                else:
                    self.logger.info("----Parsing the Config File----")
                    userCred ={"host":configur.get(self.configData["SOURCE_DB"],'host'),"dbName":configur.get(self.configData["SOURCE_DB"],'databaseName'),"userName":configur.get(self.configData["SOURCE_DB"],'userName'),"password":configur.get(self.configData["SOURCE_DB"],'password'),"port":configur.get(self.configData["SOURCE_DB"],'port')}
                    targetCred ={"host":configur.get(self.configData["TARGET_DB"],'host'),"dbName":configur.get(self.configData["TARGET_DB"],'databaseName'),"userName":configur.get(self.configData["TARGET_DB"],'userName'),"password":configur.get(self.configData["TARGET_DB"],'password'),"port":configur.get(self.configData["TARGET_DB"],'port')} 
                    self.reporter.addRow("Parsing DB Conf","Parsing of DB config is Successfull",status.PASS)
            except Exception as e:
                self.logger.error(str(e))
                traceback.print_exc()
                self.reporter.addRow("Parsing DB Conf","Exception Occurred",status.FAIL)
                output = writeToReport(self)
                self.addReasonOfFailure(traceback)
                return output, None
            self.keys = self.configData["KEYS"].split(',')
            self.reporter.addMisc("KEYS",", ".join(self.keys))
            # parsing row and connection row
            self.li1 = []
            self.li2 = []
            for i in self.keys:
                column.append(i)
                self.li1.append([])
                self.li2.append([])
            
            """connecting to sourceDB"""
            try:
                self.logger.info("----Connecting to SourceDB----")
                if self.configData["DATABASE"].lower() == 'custom':
                    db = self.configData['SOURCE_CONN']
                    ind = db.index("(")
                    module = db[0:ind-8:1]
                    __import__(module)
                    myDB = eval(db)
                elif self.configData["DATABASE"].lower() == 'mysql':
                    myDB = mysql.connector.connect(host= userCred['host'],user = userCred["userName"], password = userCred['password'],database= userCred['dbName'], port = userCred['port'])
                elif self.configData["DATABASE"].lower() == 'postgresql':
                    myDB = pg8000.connect(host= userCred['host'],user = userCred["userName"], password = userCred['password'],database= userCred['dbName'], port = userCred['port'])
                self.reporter.addRow("Connection to SourceDB: "+ str(userCred["host"]),"Connection to SourceDB is Successfull",status.PASS)
                myCursor = myDB.cursor()
            except Exception as e:
                self.logger.error(str(e))
                self.reporter.addRow("Connection to SourceDB: "+ str(userCred["host"]),"Exception Occurred",status.FAIL)
                self.addReasonOfFailure(traceback)
                
            try:
                self.logger.info("----Executing the SourceSQL----")
                myCursor.execute(self.configData['SOURCE_SQL'])
                self.reporter.addRow("Executing Source SQL","Source SQL executed Successfull",status.PASS)
                sourceColumns = [i[0] for i in myCursor.description]
            except Exception as e:
                self.logger.error(str(e))
                self.reporter.addRow("Executing Source SQL","Exception Occurred",status.FAIL)
                output = writeToReport(self)
                self.addReasonOfFailure(traceback)
                return output, None
            sorKeys =[]
            try:
                for i in self.keys:
                    if i in sourceColumns:
                        continue
                    else:
                        sorKeys.append(i)
                if len(sorKeys) == 0:
                    self.reporter.addRow("Matching Given Keys in Source SQL","Keys are Present in SourceDB",status.PASS)
                else:
                    raise Exception
            except Exception:
                keyString1 = ", ".join(sorKeys)
                self.reporter.addRow("Matching Given Keys in Source SQL","Keys: " + keyString1 +" are not Present in SourceDB",status.FAIL)
                self.logger.info("------Given Keys are not present in DB------")
                self.addReasonOfFailure(traceback)
            results = myCursor.fetchall()
            db_1 = pd.DataFrame(results, 
                                columns=sourceColumns)
            myDB.close()
            
            """Connecting to TargetDB"""
            try:
                self.logger.info("----Connecting to TargetDB----")

                if self.configData["DATABASE"].lower() == 'custom':
                    db = self.configData['TARGET_CONN']
                    ind = db.index("(")
                    module = db[0:ind-8:1]
                    __import__(module)
                    myDB1 = eval(db)
                elif self.configData["DATABASE"].lower() == 'mysql':
                    myDB1 = mysql.connector.connect(host= targetCred['host'],user = targetCred["userName"], password = targetCred['password'],database= targetCred['dbName'], port= targetCred['port'])
                elif self.configData["DATABASE"].lower() == 'postgresql':
                    myDB1 = pg8000.connect(host= targetCred['host'],user = targetCred["userName"], password = targetCred['password'],database= targetCred['dbName'],port= targetCred['port'])
                self.reporter.addRow("Connection to TargetDB: "+ str(targetCred['host']),"Connection to TargetDB is Successfull",status.PASS)
                myCur = myDB1.cursor()
            except Exception as e:
                self.logger.error(str(e))
                self.reporter.addRow("Connection to TargetDB: "+ str(targetCred["host"]),"Exception Occurred",status.FAIL)
                self.addReasonOfFailure(traceback)
            try:
                self.logger.info("----Executing the TargetSQL----")
                myCur.execute(self.configData['TARGET_SQL'])
                self.reporter.addRow("Executing Target SQL","Source SQL executed Successfull",status.PASS)
                targetColumns = [i[0] for i in myCur.description]
            except Exception as e:
                self.logger.error(str(e))
                self.reporter.addRow("Executing Target SQL","Exception Occurred",status.FAIL)
                output = writeToReport(self)
                self.addReasonOfFailure(traceback)
                return output, None
            tarKeys =[]
            try:
                for i in self.keys:
            
                    if i in targetColumns:
                        continue
                    else:
                        tarKeys.append(i)
                if len(tarKeys)==0:
                    self.reporter.addRow("Matching Given Keys in Target SQL","Keys are Present in TargetDB",status.PASS)
                else:
                    raise Exception
            except Exception:
                    keyString2 = ", ".join(tarKeys)
                    self.reporter.addRow("Matching Given Keys in Target SQL","Keys: " + keyString2 +" are not Present in TargetDB",status.FAIL)
                    self.addReasonOfFailure(traceback)
            try:
                if sourceColumns==targetColumns:
                    pass
                else:
                    raise Exception
            except Exception:
                self.reporter.addRow("Column in Table","Not Found",status.FAIL)
                self.logger.info("--------Same Column not Present in Both Table--------")
                output = writeToReport(self)
                self.addReasonOfFailure(traceback)
                return output, None
            result1 = myCur.fetchall()
            db_2 = pd.DataFrame(result1, 
                                columns=targetColumns)
            
    
            self.df_compare(db_1, db_2, self.keys)
            self.reporter.finalizeReport()
            output = writeToReport(self)
            return output, None
        except Exception as e:
            self.logger.error(str(e))
            traceback.print_exc()
            self.addReasonOfFailure(traceback)


    def setVars(self):
        """
        For setting variables like testcase name, output folder etc.
        """
        self.default_report_path = os.path.join(os.getcwd(), "pyprest_reports")
        self.data["OUTPUT_FOLDER"] = self.data.get("OUTPUT_FOLDER", self.default_report_path)
        if self.data["OUTPUT_FOLDER"].strip(" ") == "":
            self.data["OUTPUT_FOLDER"] = self.default_report_path
        self.project = self.data["PROJECT_NAME"]
        self.tcname = self.data["config_data"]["NAME"]
        self.env = self.data["ENV"]
        self.category = self.data["config_data"].get("CATEGORY", None)

    def df_compare(self,df_1, df_2, key):
        
        try:
            self.df_1 = df_1
            self.df_2 = df_2
            self.df_1['key'] = self.df_1[key].apply(lambda row: '--'.join(row.values.astype(str)), axis=1)
            self.df_2['key'] = self.df_2[key].apply(lambda row: '--'.join(row.values.astype(str)), axis=1)
            self.df_1.set_index('key', inplace=True)
            self.df_2.set_index('key', inplace=True)
            self.df_1.drop(key, axis=1, inplace=True)
            self.df_2.drop(key, axis=1, inplace=True)
            src_key_values = self.df_1.index.values
            tgt_key_values = self.df_2.index.values
            self.headers = list(set(list(df_1.columns)) - set(key+['key']))
            self.common_keys = list(set(src_key_values) & set(tgt_key_values))
            self.keys_only_in_src = list(set(src_key_values) - set(tgt_key_values))
            self.keys_only_in_tgt = list(set(tgt_key_values) - set(src_key_values))

            self.addExcel()
    
        except Exception:
            traceback.print_exc()
            self.addReasonOfFailure(traceback)
    def truncate(self,f, n):
        return math.floor(f * 10 ** n) / 10 ** n
    def addExcel(self):
        try:
            
            dict1 = { 'REASON OF FAILURE':[]}
            dict3 = {}
            dict2 = { 'Column_Name':[],'Source_Value':[],'Target_Value':[],'REASON OF FAILURE':[]}
            dict4 ={}
            outputFolder = self.data['OUTPUT_FOLDER'] + "\\"
            excelPath = outputFolder + self.configData['NAME'] + '.xlsx'
            excel = ".\\testcases\\" + self.configData['NAME'] + '.xlsx'
            keysCheck =0

            if self.keys_only_in_src:
            
                for i in self.keys_only_in_src:
            
                    li = i.split("--")
                    for i in range(len(li)):
                        key = self.keys[i]
                        self.li1[i].append(li[i])
                        value = self.li1[i]
                        dict3[key] = value
                    dict1.get("REASON OF FAILURE").append("keys only in source")
                    keysCheck +=1
            if self.keys_only_in_tgt:
                for i in self.keys_only_in_tgt:
                    li = i.split("--")
                    for i in range(len(li)):
                        key = self.keys[i]
                        self.li1[i].append(li[i])
                        value = self.li1[i]
                        dict3[key] = value
                    dict1.get("REASON OF FAILURE").append("keys only in target")
                    keysCheck +=1
            valueCheck =0
            if self.common_keys:
                
                for key_val in self.common_keys:
                    for field in self.headers:
                        src_val = self.df_1.loc[key_val,field]
                        tgt_val = self.df_2.loc[key_val,field]
                        if "THRESHOLD" in self.configData:
                            self.reporter.addMisc("Threshold",str(self.configData["THRESHOLD"]))
                            if type(src_val)== numpy.float64 :
                                src_val = self.truncate(src_val,int(self.configData["THRESHOLD"]))
                            if type(tgt_val)== numpy.float64:
                                tgt_val = self.truncate(tgt_val,int(self.configData["THRESHOLD"]))

                        if src_val != tgt_val and type(src_val)==type(tgt_val):
                            valueCheck +=1
                            li = key_val.split('--')
                        
                            for i in range(len(li)):
                                key = self.keys[i]
                                self.li2[i].append(li[i])
                                value = self.li2[i]
                                dict4[key] = value
                            dict2.get('Column_Name').append(field)
                            dict2.get('Source_Value').append(src_val)
                            dict2.get('Target_Value').append(tgt_val)
                            dict2.get('REASON OF FAILURE').append("Difference In Value")
                    
                        elif type(src_val)!= type(tgt_val):

                            valueCheck +=1
                            li = key_val.split('--')
                            for i in range(len(li)):
                                key = self.keys[i]
                                self.li2[i].append(li[i])
                                value = self.li2[i]
                                dict4[key] = value
                            dict2.get('Column_Name').append(field)
                            dict2.get('Source_Value').append(src_val)
                            dict2.get('Target_Value').append(tgt_val)
                            dict2.get('REASON OF FAILURE').append("Difference In Datatype")
            dict3.update(dict1)
            dict4.update(dict2)
            df1_res = pd.DataFrame(dict3)
            df2_res = pd.DataFrame(dict4)
            if (keysCheck+valueCheck) == 0:
                self.reporter.addRow("Data Validation Report","No MisMatch Value Found", status= status.PASS )
                self.logger.info("----No MisMatch Value Found----")
            else:
                self.logger.info("----Adding Data to Excel----")
                with pd.ExcelWriter(excelPath) as writer1:
                    if keysCheck == 0:
                        pass
                    else:
                        df1_res.to_excel(writer1, sheet_name = 'key_difference', index = False)
                    if valueCheck ==0:
                        pass
                    else:
                        df2_res.to_excel(writer1, sheet_name = 'value_difference', index = False)
                self.reporter.addRow("Data Validation Report","DVM Result File: "+'<a href='+excel+'>Result File</a>', status= status.FAIL )
            self.reporter.addMisc("common Keys", str(len(self.common_keys)))
            self.reporter.addMisc("Keys Only in Source",str(len(self.keys_only_in_src)))
            self.reporter.addMisc("Keys Only In Target", str(len(self.keys_only_in_tgt)))

        except Exception as e :
            self.logger.error(str(e))
            self.reporter.addRow("Data Validation Report","Exception Occurred", status.FAIL)
            self.addReasonOfFailure(traceback)

    def addReasonOfFailure(self,rof):
        exceptiondata = rof.format_exc().splitlines()
        exceptionarray = [exceptiondata[-1]] + exceptiondata[1:-1]
        self.reporter.addMisc("reason of failure",exceptionarray[0])



import os
import tempfile
import json
import configparser
import psutil
import sys
from gempyp.engine import dataUpload
from gempyp.reporter.reportGenerator import TemplateData
import logging


def pid_running(pid):
    if not pid:
        return False
    if psutil.pid_exists(int(pid)):
        return True
    else:
        return False

def create_report(data, s_run_id):

    data = json.loads(data)
    json_data = data[s_run_id]
    testcaseData = data["testcases"]
    output_folder = data["OUTPUT_FOLDER"]
    repJson, ouput_file_path = TemplateData().makeSuiteReport(json.dumps(json_data), testcaseData, output_folder)
    suite_data = json_data["suits_details"]
    suite_data["s_id"] = data["s_id"]
    suite_data["misc_data"] = data["misc_data"]
    username = suite_data["initiated_by"]
    del suite_data["testcase_details"]
    del suite_data["Testcase_Info"]

    # read bridgetoken
    try:
        config_file = configparser.ConfigParser()
        config_file.read("gempyp.conf")
        bridgetoken = config_file['ReportSetting'].get("bridge_token", None)
    except Exception as e:
        bridgetoken = None
    dataUpload.sendSuiteData(json.dumps(suite_data), bridgetoken, username, mode="PUT")
    return ouput_file_path


if __name__ == "__main__":
    while(True):
        if not pid_running(os.environ["PID"]):
            s_run_id = os.getenv("S_RUN_ID")
            if s_run_id:
                logging.basicConfig(level=logging.DEBUG)
                file_path = os.path.join(tempfile.gettempdir(), s_run_id + ".txt")
                if os.path.exists(file_path):
                    with open(file_path, "r+") as f:
                        data = str(f.read())
                        output_file_path = create_report(data, s_run_id)
                        # where to get this json data from
                        current_pid = os.getpid()
                        logging.info(f"Find Gempyp logs at - {s_run_id + '.log'}")
                        logging.info(f"Find Gempyp Report at - {output_file_path}")
                        os.rename("logs.log",f"{s_run_id}.log")
                        os.kill(current_pid, 19)
                        sys.exit()



from gempyp.engine.simpleTestcase import AbstarctSimpleTestcase
from gempyp.engine.baseTemplate import TestcaseReporter


class Gempyp(AbstarctSimpleTestcase):
    def __init__(self, project_name=None, testcase_name=None):
        if project_name is not None and testcase_name is not None:
            self.reporter = TestcaseReporter(project_name, testcase_name)
        else:
            self.reporter = TestcaseReporter("Gempyp", "testing")
        
        # access project name and testcase name
        # self.reporter = TestcaseReporter(Project, Testcase)
        # user will use this reporter in their testcases
        


        


import cx_Oracle
import os
import sys
import traceback
import time
import common.common_functions as cf

class Database():
    def __init__(self, config_data) -> None:
        try:
            self.user = config_data.get('user',None)
            self.password = config_data.get('password',None)
            self.database = config_data.get('database',None)
            self.port = config_data.get('port',None)
            self.dsn = config_data.get('dsn',None)
        except Exception as exc:
            print("Exception in __init__ of db_oracle.Database {}".format(str(exc)))
    
    def connect(self,threaded = False):
        try:
            if not self.dsn:
                if self.user and self.password and self.database and self.port:
                    self.dsn = self.user + "/" +cf.decrypt(self.password) +"@" + self.database+":"+self.port
                elif self.database:
                    self.dsn = "/@"+self.database
            if threaded:
                self.con = cx_Oracle.connect(self.dsn,threaded=True)
            else:
                self.con = cx_Oracle.connect(self.dsn)
            print("Connected to db successfully")
            self.cursor = self.con.cursor()
            self.cursor.arraysize = 1000
            print("cursor created successfully")
        except Exception:
            print("Exception occured in connect {}".format(traceback.format_exc()))

    def execute_query(self, query, header = True):
        """
        We will be only executing select from this so we are adding header as param
        """
        try:
            self.cursor.execute(query)
            res = self.cursor.fetchall()
            if header:
                q_header = [x[0] for x in self.cursor.description]
            else:
                q_header = []

        except Exception:
            res = None
            q_header = None
            print("Exception in execute_query {}".format(traceback.format_exc()))
        return res,q_header if header else res

    def execute_dm_query(self, query):
        try:
            r_val = 1
            self.cursor.execute(query)
            self.con.commit()
            r_val = 0
        except Exception:
            self.con.rollback()
            print("Exception occured in execute_dml_query {}".format(traceback.format_exc()))
        return r_val
    
    def close(self):
        try:
            print("CLosing db connection")
            self.con.close()
            print("Db connection closed")
        except Exception:
            pass

if __name__ == "__main__":
    pass



from asyncio.log import logger
import traceback
import os
import uuid
import time
import json
import logging
import getpass


def writeToReport(pyprest_obj):
    """
    This function takes pyprest obj,
    creates the testcase report if the runmode is debug mode,
    Creates the dictionary that is to be sent to gempyp"""
    result = {}
    if not pyprest_obj.reporter.result_file_name:
        try:
            try:
                pyprest_obj.reporter.finalizeReport()   ## need to test
                if pyprest_obj.data.get("OUTPUT_FOLDER", pyprest_obj.default_report_path) is None:
                    os.makedirs(pyprest_obj.data.get("OUTPUT_FOLDER", pyprest_obj.default_report_path))
            except Exception as e:
                pyprest_obj.logger.info(traceback.print_exc())
            pyprest_obj.reporter.json_data = pyprest_obj.reporter.template_data.makeTestcaseReport()
            pyprest_obj.json_data = pyprest_obj.reporter.json_data
            result = pyprest_obj.reporter.serialize()
            if pyprest_obj.data["config_data"].get("DEBUG_MODE", "FALSE").upper() == "TRUE":
                # make report
                try:
                    makeReport_pypRest(pyprest_obj, json.dumps(pyprest_obj.reporter.json_data))
                    pyprest_obj.logger.info("-------file_dumped---------")
                except Exception as e:
                    pyprest_obj.logger.info(str(e))

        except Exception as e:
            pyprest_obj.logger.info(traceback.print_exc())

    output = []
    tempdict = {} 
    unique_id = uuid.uuid4()
    try:
        if os.environ.get('unique_id'):
            unique_id = os.environ.get('unique_id')
    except Exception:
        traceback.print_exc()
    tc_run_id = f"{pyprest_obj.tcname}_{unique_id}"
    tempdict["tc_run_id"] = tc_run_id
    tempdict["name"] = result["NAME"]
    tempdict["category"] = pyprest_obj.category
    tempdict["status"] = result["STATUS"]
    tempdict["base_user"] = getpass.getuser()
    tempdict["invoke_user"] = pyprest_obj.data.get("INVOKE_USER")
    tempdict["machine"] = pyprest_obj.data.get("MACHINE")
    tempdict["product_type"] = "GEMPYP-PR"
    tempdict["result_file"] = result["RESULT_FILE"]
    tempdict["start_time"] = result["START_TIME"]
    tempdict["end_time"] = result["END_TIME"]
    tempdict["ignore"] = False  # to be updated
    all_status = result["json_data"]["meta_data"][2]
    total = 0
    for key in all_status:
        total += all_status[key]
    result["json_data"]["meta_data"][2]["TOTAL"] = total

    # getting the log file ( the custom gempyp logger)
    
    tempdict["log_file"] = pyprest_obj.data.get("LOG_PATH", "N.A")

    singleTestcase = {}
    singleTestcase["testcase_dict"] = tempdict
    singleTestcase["misc"] = result.get("MISC")
    singleTestcase["json_data"] = pyprest_obj.json_data
    singleTestcase["suite_variables"] = pyprest_obj.variables.get("suite", {})
    output.append(singleTestcase)

    return output


def makeReport_pypRest(obj, json_data):
        # Create testcase file in the given output folder when in debug mode

        index_path = os.path.dirname(__file__)
        result_data = ""
        index_path = os.path.join(os.path.split(index_path)[0], "testcase.html")
        with open(index_path, "r") as f:
            result_data = f.read()

        result_data = result_data.replace("::DATA::", json_data)

        result_file = os.path.join(obj.data.get("OUTPUT_FOLDER"), f"{obj.reporter.testcase_name + str(time.time())}.html")
        with open(result_file, "w+") as f:
            f.write(result_data)


import os
import json
import logging
from datetime import datetime
import traceback
from typing import List, Union
import typing
from gempyp.config import DefaultSettings
import importlib
import sys


def read_json(file_path):
    try:
        with open(file_path) as fobj:
            res = json.load(fobj)
            print(res)
    except Exception:
        res = None
    return res
def readPath(file_name):
        try:
            conf = file_name.split(os.sep)
            if conf[0] == ".." or conf[0]==".":
                script_path, script_name = importFromPath(file_name)
                for each in sys.path:
                    if isinstance(each,dict) and each is not None:
                        logging.info("--------- Fetching config path - {} ------".format( each['XMLConfigDir']))
                        lib_path = os.path.join(each['XMLConfigDir'], file_name)
                        return lib_path
            else:
                return file_name
        except Exception as e:
            traceback.print_exc()


def findDuration(start_time: datetime, end_time: datetime):
    # finds the duration in the form HH MM SS

    duration = end_time - start_time
    seconds = duration.total_seconds()
    mins = seconds // 60
    seconds = seconds % 60

    if mins > 0:
        return f"{mins} mins and {round(seconds, 3)} seconds"
    return f"{round(seconds, 3)} seconds"


def errorHandler(logging, Error, msg="some Error Occured"):

    logging.error(msg)
    logging.error(f"Error: {Error}")

    if DefaultSettings.DEBUG:
        logging.error(f"TraceBack: {traceback.format_exc()}")


def parseMails(mail: Union[str, typing.TextIO]) -> List:
    try:

        if hasattr(mail, "read"):
            mails = mail.read()

        elif os.path.isfile(mail):
            file = open(mail, "r")
            mails = file.read()
            file.close()

        mails = mail.strip()
        mails = mails.split(",")
        return mails
    except Exception as e:
        logging.error("Error while parsing the mails")
        logging.error(f"Error : {e}")
        logging.error(f"traceback: {traceback.format_exc()}")
        return []


# custom encoder to encode date to epoch
class dateTimeEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.timestamp()*1000
        return json.JSONEncoder.default(self, o)

# absolute path path.... moved to file from runner due to circular import issue
# runner -> gempyp -> abstractsimpletestcase -> runner
def importFromPath(file_name):
    logging.info("--------- In import from path ----------")
    path_arr = file_name.split(os.sep)    
    file = path_arr[-1]
    path_arr.remove(file)
    path_cd = os.sep.join(path_arr)
    return path_cd, file
    
def moduleImports(file_name):
    import_flag = 0
    try:
        logging.info("--------Trying importing modules--------")
        dynamicTestcase = importlib.import_module(file_name)        
        return dynamicTestcase
    except Exception as i:
        logging.info("-------Testcase not imported as module, Trying with absolute path-------")
        try:
            script_path, script_name = importFromPath(file_name)
            script_name = script_name.split(".")[0]
            if script_path is not None:
                sys.path.append(script_path)
            try:
                dynamicTestcase = importlib.import_module(script_name) 
            except Exception:
                logging.info("when absolute and module import both failed")
                try:
                    for each in sys.path:
                        if isinstance(each,dict) and each is not None:
                            logging.info("--------- Fetching config path - {} ------".format( each['XMLConfigDir']))
                            lib_path = os.path.join(each['XMLConfigDir'], script_path)
                    sys.path.append(lib_path)
                    dynamicTestcase = importlib.import_module(script_name.split(".")[0])
                    import_flag = 1 
                except Exception:
                    traceback.print_exc()
                if import_flag != 1:
                    traceback.print_exc()
            return dynamicTestcase
        except Exception as e:
            logging.error("----- Error occured file could not be imported using any of the methods.-----")
            traceback.print_exc()
            return e




import tempfile
from typing import Dict
import lxml.etree as et
import logging
from gempyp.config.baseConfig import AbstarctBaseConfig
from gempyp.libs.parsers import xmlToDict, xmlToList
from gempyp.libs.logConfig import LoggingConfig
import sys, os
import warnings
import uuid

class XmlConfig(AbstarctBaseConfig):
    def __init__(self, filePath: str, s_run_id):
        
        # code pushed in parse function to make run_id working
        self.s_run_id = s_run_id
        super().__init__(filePath)
        # do any xml specific validatioins here
       

    def parse(self, filePath):

        logging.info("-------- Xml file path: {filePath} ----------".format(filePath=filePath))
        path_list = filePath.split(os.sep)[0:-1]
        newfilePath = os.sep.join(path_list)
        sys.path.append({"XMLConfigDir":newfilePath})
        logging.info("-------- Started the Xml parsing in XmlConfig ---------")
        self.handleSpecialSymbols(filePath)
        data = et.parse(filePath)
        self._CONFIG["SUITE_DATA"] = self._getSuiteData(data)        

        self.log_dir = str(os.path.join(tempfile.gettempdir(), 'logs'))
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)
        self.unique_id = str(uuid.uuid4())
        if self.s_run_id != None:
             self.unique_id = self.s_run_id
        elif "RUN_ID" in self._CONFIG["SUITE_DATA"]:
            self.unique_id = self._CONFIG["SUITE_DATA"]["RUN_ID"]

        os.environ['unique_id'] = self.unique_id
        os.environ['log_dir'] = self.log_dir
        warnings.filterwarnings('ignore')
        suiteLogsLoc = str(os.path.join(self.log_dir, 'Suite_' + self.unique_id + '.log'))
        LoggingConfig(os.path.join(self.log_dir, 'Suite_' + self.unique_id + '.log'))
        # LoggingConfig(suiteLogsLoc)
        print("------suite logs------",suiteLogsLoc)


        self._CONFIG["TESTCASE_DATA"] = self._getTestCaseData(data)

        self._CONFIG["SUITE_DATA"]['LOG_DIR'] = self.log_dir
        self._CONFIG["SUITE_DATA"]['UNIQUE_ID'] = self.unique_id

    def _getSuiteData(self, data) -> Dict:

        suite_data = data.find("suite")

        suite_dict = xmlToDict(suite_data)
        suite_dict["SUITE_VARS"] = {}
        logging.info("--------suite_dict--------\n {suite_dict} \n----------".format(suite_dict=suite_dict))
        # do your validations here

        return suite_dict

    def _getTestCaseData(self, data) -> Dict:

        testcase_data = data.find("testcases")

        testcase_list = xmlToList(testcase_data)

        for i in testcase_list:
            i["NAME"] = i["NAME"].upper()  ## uppercase
        testcase_dict = {k['NAME']: k for k in testcase_list}  #####################

        # do your validation here

        return testcase_dict
    
    def handleSpecialSymbols(self,filePath):
        f=open(filePath,"r")
        content=f.read()
        if(content.__contains__("&") and not(content.__contains__("&amp;"))):
            content=content.replace("&","&amp;")
        f1=open(filePath,"w")
        f1.write(content)



from fileinput import filename
from stat import filemode
import sys, os
# from tkinter import W
import coloredlogs, logging

class LoggingConfig:
    def __init__(self, file_name='tmp.log') -> None:
        # logging.basicConfig(filename=file_name,
        #             format='%(asctime)s [%(levelname)s] - [%(filename)s > %(funcName)s() > %(lineno)s] - %(message)s',
        #             datefmt='%H:%M:%S',
        #             filemode='w',
        #             level=logging.INFO)
        self.mylogs = logging.getLogger()
        fieldstyle = {'asctime': {'color': 'green'},
                    'levelname': {'bold': True, 'color': 'white'},
                    'filename':{'color':'cyan'},
                    'funcName':{'color':'blue'}}
                                        
        levelstyles = {'critical': {'bold': True, 'color': 'red'},
                    'debug': {'color': 'green'}, 
                    'error': {'color': 'red'}, 
                    'info': {'color':'white'},
                    'warning': {'color': 'yellow'}}

        coloredlogs.install(level=logging.INFO,
                            logger=self.mylogs,
                            fmt='%(asctime)s [%(levelname)s] - [%(filename)s > %(funcName)s() > %(lineno)s] - %(message)s',
                            datefmt='%H:%M:%S',
                            field_styles=fieldstyle,
                            level_styles=levelstyles)
        file = logging.FileHandler(filename=file_name, mode='w')
        fileformat = logging.Formatter('%(asctime)s [%(levelname)s] - [%(filename)s > %(funcName)s() > %(lineno)s] - %(message)s',
        datefmt='%d-%m-%Y, %H:%M:%S')
        file.setFormatter(fileformat)
        self.mylogs.addHandler(file)

def my_custom_logger(logger_name, level=logging.INFO):
    """
    Method to return a custom logger with the given name and level
    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)
    format_string = '%(asctime)s [%(levelname)s] - [%(filename)s > %(funcName)s() > %(lineno)s] - %(message)s'
    log_format = logging.Formatter(format_string, datefmt='%d-%m-%Y, %H:%M:%S')
    # Creating and adding the console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(log_format)
    logger.addHandler(console_handler)
    # Creating and adding the file handler
    file_handler = logging.FileHandler(logger_name, mode='w')
    file_handler.setFormatter(log_format)
    logger.addHandler(file_handler)
    return logger
        



import boto3
import os


def download(bucket=None, file_name=None, folder=None):
    try:
        s3 = boto3.resource('s3')
        my_bucket = s3.Bucket(bucket)
        for object_summary in my_bucket.objects.filter(Prefix="/".join(folder)):
            body = object_summary.get()['Body'].read()

        fileContent = body.decode().split("\\n")
        path = os.path.join(file_name)
        with open(path, "w+") as fp:
            fp.seek(0)
            fp.write('\n'.join(fileContent))
            fp.truncate()

        return path
    except Exception as e:
        print(str(e))
        return f"error occurred- {str(e)}"


from asyncio.log import logger
import getpass
import json
from pathlib import Path
import sys
import uuid
from datetime import datetime
from gempyp.engine.baseTemplate import TestcaseReporter
from gempyp.engine.runner import getOutput
from gempyp.engine.testData import TestData
from gempyp.engine import dataUpload
import configparser
import inspect
from datetime import timezone
import platform
import logging
import tempfile
import subprocess
import sys
import os
import pandas as pd
from gempyp.libs.enums.status import status


class Executor(TestcaseReporter):
    def __init__(self, **kwargs):
        self.method = kwargs.get("tc_name", self.getMethodName())
        self.log_file = tempfile.gettempdir() + "\logs.log"
        # os.makedirs("testcase_log_folder")
        sys.stdout = sys.stderr =  open(self.log_file, 'w')
        logging.basicConfig(filename="logs.log", filemode='w', format='%(name)s - %(asctime)s-%(levelname)s - %(message)s',level=logging.DEBUG)
        # custom_logger = my_custom_logger("logs.log")
        logging.info("inside constructor here--------------------")
        logging.info(f"-------Executing testcase - {self.getMethodName()}---------")
        self.data = self.getTestcaseData()
        self.reporter = TestcaseReporter(self.data["PROJECT"], self.data["NAME"])
       
        

        path = __file__
        path = path.rsplit(os.sep, 1)[0]
        self.DATA = TestData()
        # make suite details and upload it
        self.makeSuiteDetails()
        if not os.getenv("PID"):
            self.makeOutputFolder()
            os.environ["PID"] = str(os.getpid())
            subprocess.Popen([sys.executable, os.path.join(path, "worker.py")], shell=True)
            try:
                logging.info(f"---------------S_RUN_ID-------------{self.s_run_id}")
                dataUpload.sendSuiteData((self.DATA.toSuiteJson()), self.data["BRIDGE_TOKEN"], self.data["USERNAME"]) # check with deamon, should insert only once
                  # logging not working
            except Exception as e:
                print(f"Exception occured - {e}")
                pass

    def __del__(self):
        self.final()

    def final(self):       
        output = []
        
        # destructor of reporter object called
        
        self.reporter.finalizeReport()
        
        # create testcase reporter json
        self.reporter.json_data = self.reporter.template_data.makeTestcaseReport()
        # serializing data, adding suite data
        report_dict = self.reporter.serialize()

        report_dict["TESTCASEMETADATA"] = self.getMetaData()
        report_dict["config_data"] = self.getConfigData()

        # creating output json
        output.append(getOutput(report_dict))
        output[0]["product_type"] = "GEMPYP-SDK"
        for i in output:
            logging.info("---------------TC_RUN_ID-----------------"+i["testcase_dict"]["tc_run_id"].upper())
            i["testcase_dict"]["steps"] = i["json_data"]["steps"]
            dict_ = {}
            dict_["testcases"] = {}
            dict_["OUTPUT_FOLDER"] = os.getenv("OUTPUT_FOLDER")
            dict_["misc_data"] = {}
            tmp_dir = os.path.join(tempfile.gettempdir(), self.s_run_id + ".txt")
            

            self.DATA.testcase_details = self.DATA.testcase_details.append(
                i["testcase_dict"], ignore_index=True
            )
            # self.DATA.testcaseDetails = pd.concat([self.DATA.testcaseDetails, pd.DataFrame(list(i["testcase_dict"].items()))])
            self.updateTestcaseMiscData(i["misc"], tc_run_id=i["testcase_dict"].get("tc_run_id"))
            suite_data = self.DATA.getJSONData()
            if isinstance(suite_data, str):
                suite_data = json.loads(suite_data)
            if isinstance(self.DATA.toSuiteJson(), str):
                suite_temp = json.loads(self.DATA.toSuiteJson())
            if not os.path.exists(tmp_dir):
                with open(tmp_dir, "w") as f:
                    dict_[self.s_run_id] = self.updateSuiteData(suite_data)
                    dict_["s_id"] = suite_temp["s_id"]
                    dict_["misc_data"] = suite_temp["misc_data"]
                    dict_["testcases"][i["testcase_dict"].get("tc_run_id")] = i["json_data"]
                    f.write(json.dumps(dict_))
            else:
                with open(tmp_dir, "r+") as f:
                    data = f.read()
                    data = json.loads(data)
                    data[self.s_run_id] = self.updateSuiteData(suite_data, data[self.s_run_id])
                    data["s_id"] = suite_temp["s_id"]
                    data["misc_data"] = suite_temp["misc_data"]
                    data["testcases"][i["testcase_dict"].get("tc_run_id")] = i["json_data"]
                    f.seek(0)
                    f.write(json.dumps(data))
            
            dataUpload.sendTestcaseData((self.DATA.totestcaseJson(i["testcase_dict"]["tc_run_id"].upper(), self.data["S_RUN_ID"])), self.data["BRIDGE_TOKEN"], self.data["USERNAME"])  # instead of output, I need to pass s_run id and  tc_run_id
            
            # sys.stdout.close()
        
            # os.rename(self.log_file, tmp_dir.rsplit(".", 1)[0] + ".log")

            

    def getTestcaseData(self):
        config_file = configparser.ConfigParser()
        directory_path = os.getcwd()

        if not os.path.exists(directory_path + os.sep + "gempyp.conf"):
            print("Config file is missing. Aborting  gempyp report......")
            sys.exit()
        config_file.read("gempyp.conf")
        data = {}
        self.projectName = data["PROJECT"] = config_file['ReportSetting']["project"]
        self.testcaseName = data["NAME"] = self.method
        self.env = data["ENV"] = config_file['ReportSetting'].get("env", "PROD")
        data["USERNAME"] = config_file['ReportSetting'].get("USERNAME", getpass.getuser())
        data["BRIDGE_TOKEN"] = config_file['ReportSetting'].get("BRIDGE_TOKEN", None)
        data["OUTPUT_FOLDER"] = config_file['ReportSetting'].get("outputfolder", None)
        data["MACHINE"] = platform.node()
        data["MAIL"] = config_file['ReportSetting'].get("mail", None)
        self.report_name = data["REPORT_NAME"] = config_file['ReportSetting'].get("reportname", "SMOKE_TEST")
        self.report_info=data["REPORT_INFO"]=config_file['ReportSetting'].get("reportinfo", "SMOKE_TEST_INFO")
        if not os.getenv("S_RUN_ID"):
            s_run_id = data["PROJECT"] + "_" + data["ENV"] + "_" + str(uuid.uuid4())
            os.environ["S_RUN_ID"] = s_run_id.upper()
        self.s_run_id = data["S_RUN_ID"] = config_file['ReportSetting'].get("s_run_id", os.getenv("S_RUN_ID"))
        return data

    def getMetaData(self):
        data = {
            'PROJECTNAME': self.data["PROJECT"], 
            'ENV': self.data["ENV"], 
            'S_RUN_ID': self.data["S_RUN_ID"], 
            'USER': self.data["USERNAME"], 
            'MACHINE': self.data["MACHINE"], 
            'OUTPUT_FOLDER': self.data["OUTPUT_FOLDER"]}
        return data

    def getMethodName(self):
        try:
            currframe = inspect.currentframe()
            callframe = inspect.getouterframes(currframe, 2)
            count = -1
            while count < 0:
                if callframe[2][3] != '<module>':
                    # method = callframe[2][3]
                    filename = callframe[2][1].split(os.sep)[-1].split(".")[0]
                    method = callframe[2][3] + "_" + filename
                    break
                count = (count - 1)%(-len(callframe))
        except Exception as e:
            print(e)
            method = f"Method_{str(uuid.uuid4())}"
        return method

    def getConfigData(self):
        data = {'NAME': self.data["NAME"], 'CATEGORY': 'External','LOGGER': ""}
        return data

    def makeSuiteDetails(self):
        """
        making suite Details 
        """

        Suite_details = {
            "s_run_id": self.data["S_RUN_ID"],
            "s_start_time": datetime.now(timezone.utc),
            "s_end_time": None,
            "status": status.EXE.name,
            "project_name": self.data["PROJECT"],
            "user": self.data["USERNAME"],
            "report_name": self.data["REPORT_INFO"],
            "framework_name": "GEMPYP",
            "env": self.data["ENV"],
            "machine": self.data["MACHINE"],
            "os": platform.system().upper(),
        }
        self.DATA.suite_detail = self.DATA.suite_detail.append(
            Suite_details, ignore_index=True
        )

        

    def updateTestcaseMiscData(self, misc, tc_run_id):
        """
        updates the misc data for the testcases
        """
        miscList = []

        for misc_data in misc:
            temp = {}
            # storing all the key in upper so that no duplicate data is stored
            temp["key"] = misc_data.upper()
            temp["value"] = misc[misc_data]
            temp["run_id"] = tc_run_id
            temp["table_type"] = "TESTCASE"
            miscList.append(temp)

        self.DATA.misc_details = self.DATA.misc_details.append(
            miscList, ignore_index=True
        )

    def makeOutputFolder(self):
        """
        to make GemPyp_Report folder 
        """

        logging.info("---------- Making output folders -------------")
        report_folder_name = f"{self.projectName}_{self.env}"
        if self.report_name:
            report_folder_name = report_folder_name + f"_{self.report_name}"
        date = datetime.now().strftime("%Y_%b_%d_%H%M%S_%f")
        report_folder_name = report_folder_name + f"_{date}"
        if self.data.get("OUTPUT_FOLDER"):
            self.ouput_folder = os.path.join(
                self.data.get("OUTPUT_FOLDER"), report_folder_name
            )
        else:
            home = str(Path.home())
            self.ouput_folder = os.path.join(
                home, "gempyp_reports", report_folder_name
            )
        os.environ["OUTPUT_FOLDER"] = self.ouput_folder

        os.makedirs(self.ouput_folder)

    def updateSuiteData(self, current_data, old_data=None):
        """
        updates the suiteData after all the runs have been executed
        """
        start_time = current_data["suits_details"]["s_start_time"]
        end_time = current_data["suits_details"]["testcase_details"][0]["end_time"]
        testcaseData = current_data["suits_details"]["testcase_details"]
        if old_data:
            testcaseData = (old_data["suits_details"]["testcase_details"]
             + current_data["suits_details"]["testcase_details"])
            start_time = old_data["suits_details"]["s_start_time"]
        statusDict = {k.name: 0 for k in status}
        for i in testcaseData:
            statusDict[i["status"]] += 1
        # get the status count of the status
        SuiteStatus = status.FAIL.name

        # based on the status priority
        for s in status:
            if statusDict.get(s.name, 0) > 0:
                SuiteStatus = s.name

        current_data["suits_details"]["status"] = SuiteStatus
        current_data["suits_details"]["testcase_details"] = testcaseData
        current_data["suits_details"]["s_start_time"] = start_time
        current_data["suits_details"]["s_end_time"] = end_time
        count = 0
        for key in list(statusDict.keys()):
            if statusDict[key] == 0:
                del statusDict[key]
            else:
                count += statusDict[key]
        statusDict["Total"] = count
        current_data["suits_details"]["Testcase_Info"] = statusDict

        return current_data



import re
import logging as logger
from inspect import getmembers, isfunction
from gempyp.pyprest.predefinedFunctions import PredefinedFunctions as prefunc


class VariableReplacement:
    def __init__(self, pyprest_obj):
        self.pyprest_obj = pyprest_obj
        if pyprest_obj.__dict__.get("variables") is not None:
            if self.pyprest_obj.variables.get("local") is not None:
                self.local_pre_variables = self.pyprest_obj.variables.get("local")
            if self.pyprest_obj.variables.get("suite") is not None:
                self.suite_pre_variables = self.pyprest_obj.variables.get("suite")
        functions_list = [x[0] for x in getmembers(prefunc, isfunction)]
        self.functions_dict = {each.lower():each for each in functions_list}
            
        self.pyprest_obj.logger.info("**************  INSIDE VARIABLE REPLACEMENT  **************")


    def replaceToNull(self,data):
        logList = []
        for k,v in data.copy().items():
            if isinstance(v,dict):
                data[k] = self.replaceToNull(v)
            elif isinstance(v, list):
                for each in v:
                    if isinstance(each,dict):
                        self.replaceToNull(each)
            elif isinstance(v,str):
                if k!="PRE_VARIABLES" and k!="pre_variables" and k!="POST_VARIABLES" and k!="post_variables"  and k != "report_misc" and k !="REPORT_MISC" and  k is not None:
                    var_list = self.checkAndGetVariables(data[k])
                    for var in var_list:
                        var_name = var
                        var_val = self.getVariableValues(var_name)
                        if var_val == "null" and "$[#" in str(data[k]):
                            newValStr = data[k].replace(var_name, "Null")
                            logger.warning("-------Value of {0} not found!!! Assigning Null".format(var_name))
                        else:
                            newValStr = data[k].replace(var_name, str(var_val))
                        del data[k]
                        data[k] = newValStr         
        return data     
        

        
    def valueNotFound(self):
        logger.info("------ assigning all variables to null those value not found -----") 
        try:
            self.replaceToNull(self.pyprest_obj.__dict__)
            self.replaceToNull(self.pyprest_obj.reporter.template_data.__dict__)
        except Exception as e:
            print(e)


    def checkAndGetVariables(self, value_str) -> list:
        if value_str is not None and type(value_str) is str:
            self.value = value_str
            pattern = r"\$\[\S[a-zA-Z0-9_.]+\]"
            regex = re.compile(pattern) 
            variables = re.findall(regex,self.value)
            return variables
       
       
    def getVariableValues(self, var_name):
        varName = var_name.strip("$[#]")
        try:
            if "SUITE.".casefold() in varName.casefold():
                varValue = self.suite_pre_variables[varName.replace(".", "_").upper()]
            else:
                varValue = self.local_pre_variables[varName]
                # suite_variables
                # varValue = self.local_pre_variables[varName]
        except:
            return "null"
        str_val = var_name.replace("$[#"+varName+"]", str(varValue))
        if "$[#" not in str_val:
            return str(str_val) 
        if "$[#" or "$" in str_val:
            return self.getVariableValues(str_val) 
    
    def getFunctionValues(self, var_name):
        func = var_name.strip("$[#]")
        func_items = func.split("(")
        func_name = func_items[0]
        params = func_items[1].strip(")")
        func_name = self.functions_dict.get(func_name.lower(), "invalid")

        try:
            val =  (prefunc + func_name)(params) if func_name != "invalid" else "null"
            return val
        except:
            return "null"
         


    def updateDataDictionary(self, data):
        for k,v in data.copy().items():
            if isinstance(v,dict):
                data[k] = self.updateDataDictionary(v)
            elif isinstance(v,str):
                if k!="PRE_VARIABLES" and k!="pre_variables" and k!="POST_VARIABLES" and k!="post_variables"  and k != "report_misc" and k !="REPORT_MISC" and  k is not None:
                    var_list = self.checkAndGetVariables(data[k])
                    for var in var_list:
                        var_name = var
                        var_val = self.getVariableValues(var_name)
                        if var_val == "null" and "$[#" in str(data[k]):
                            newValStr = data[k]
                        else:
                            newValStr = data[k].replace(var_name, str(var_val))
                        del data[k]
                        data[k] = newValStr

        return data

    def variableReplacement(self):
        self.updateDataDictionary(self.pyprest_obj.__dict__)
        



from abc import ABC, abstractmethod
from typing import List, Union, Dict
from gempyp.config import DefaultSettings
from gempyp.engine.baseTemplate import TestcaseReporter
from gempyp.libs.common import moduleImports
import sys,traceback
from gempyp.libs.enums.status import status
import logging


class AbstractSimpleTestcase(ABC):
    Status = status
    # get logger

    def gempypMethodExecutor(
        self, cls, testcase_settings: Dict, **kwargs
    ) -> TestcaseReporter:
        """
        extend the baseTemplate and implement this method.
        :param testcase_settings: testcasesettings object created from the testcase config
        :return
        """
        logger = testcase_settings.get("LOGGER")
        kwargs["TESTCASENAME"] = testcase_settings["NAME"]
        file_name = testcase_settings.get("PATH")
        try:
            testcase = moduleImports(file_name)
        except:
            logger.ERROR("Testcase not imported")
        try:
            method_name = testcase_settings.get("METHOD", "main")
            logger.info(f"-------- method ---------{method_name}")
            reporter = TestcaseReporter(kwargs["PROJECT_NAME"], testcase_settings["NAME"])
            # adding logger to the reporter
            reporter.logger = logger
            try:
                method_name = getattr(cls(), method_name)
                method_name(reporter)
            except Exception as err:
                logger.error(traceback.format_exc())
                etype, value, tb = sys.exc_info()
                info, error = traceback.format_exception(etype, value, tb)[-2:]
                # code for finding exception and writing reason of failure 
                exceptiondata = traceback.format_exc().splitlines()
                exceptionarray = [exceptiondata[-1]] + exceptiondata[1:-1]
                reporter.addMisc("Reason of Failure",exceptionarray[0])
                #reports = TestcaseReporter(kwargs["PROJECT_NAME"], testcase_settings["NAME"])
                reporter.addRow("Exception Occured", str(error) + 'at' + str(info), status.ERR)
            finally:
                return reporter
                
        except Exception as e:
            if e:            
                logger.error(traceback.format_exc())
                exceptiondata = traceback.format_exc().splitlines()
                exceptionarray = [exceptiondata[-1]] + exceptiondata[1:-1]
                reporter.addMisc("Reason of Failure",exceptionarray[0])
            
        

    def RUN(self, cls, testcase_settings: Dict, **kwargs) -> List:
        """
        the main function which will be called by the executor
        take argument as testcaseDict and call the gempypMethodExecutor() method
        and later on call basetemplate finalizereport method and reurn data to testcaseRunner method 
        """
        # set the values from the report if not s et automatically
        self.logger = testcase_settings.get('LOGGER')
        Data = []

        try:
            self.logger.info('================= Running Testcase: {testcase} ============'.format(testcase=testcase_settings["NAME"]))
            reports = self.gempypMethodExecutor(cls, testcase_settings, **kwargs)

            # will never enter this  block
            # if reports is None:
            #     reports = TestcaseReporter(kwargs["PROJECT_NAME"], testcase_settings["NAME"])
            #     self.logger.error("Report object was not returned from the testcase file")
            #     reports.addRow("Exception Occured", "Exception occured in testcase: Report was not generated.", status.FAIL)    
        except Exception:
            etype, value, tb = sys.exc_info()
            self.logger.error(traceback.format_exc())
            info, error = traceback.format_exception(etype, value, tb)[-2:]
            reports = TestcaseReporter(kwargs["PROJECT_NAME"], testcase_settings["NAME"])
            reports.addRow("Exception Occured", str(error) + 'at' + str(info), status.ERR)


        if isinstance(reports, TestcaseReporter):
            reports = [reports]

        for index, report in enumerate(reports):
            if not report.project_name:
                report.project_name = testcase_settings.get("PROJECTNAME", "GEMPYP")

            if not report.testcase_name:
                report.testcase_name = testcase_settings.get("NAME", "TESTCASE")
                report.testcase_name = f"{self.testcase_name}_{index}"

            # call the destructor if not already called.
            report.finalizeReport()
            # if user has not provided its own resultfile
            if not report.result_file_name:
                report.json_data = report.template_data.makeTestcaseReport()
            result = report.serialize()
            Data.append(result)

        return Data


import psycopg2
import os
import sys
import traceback
import time
import common.common_functions as cf

class Database():

    
    def __init__(self, config_data) -> None:
        self.user = config_data.get("user",None)
        self.host = config_data.get("host", None)
        self.port = config_data.get("port", None)
        if isinstance(self.port,str):
            self.port = int(self.port)
        self.password = config_data.get("password", None)
        self.database = config_data.get("database", None)
        
    def connect(self):
        try:
            self.connection = psycopg2.connect(host = self.host, user = self.user, password = cf.decrypt(self.password), port = self.port,database = self.database)
            print("Connected Successfully")
            self.cursor = self.connection.cursor()
            print("Cursor Created")
        except Exception:
            print("Exception occurred in connect {}".format(traceback.format_exc()))
           
    def execute_query(self, query, header = False):
        try:
            self.cursor.execute(query)
            res = self.cursor.fetchall()
            if(header):
                q_header = [x[0] for x in self.cursor.description]
            else:
                q_header = []
        except Exception:
            res = None
            q_header = None
            print("Exception occurred in execute_query {}".format(traceback.format_exc()))
        return res, q_header if header else res
        
    def execute_dm_query(self, query):
        try:
            r_val = 1
            self.cursor.execute(query)
            self.connection.commit()
            r_val = 0
        except Exception:
            self.connection.rollback()
            print("Exception occured in execute_dml_query {}".format(traceback.format_exc()))
        return r_val
    
    def commit(self):
        try:
            self.connection.commit()
            print("Committed")
        except Exception:
            print("Exception occurred in commit {}".format(traceback.format_exc()))
        
    def close(self):
        try:
            #self.cursor.close()
            self.connection.close()
            print("Connection Closed")
        except Exception:
            print("Exception occurred in close {}".format(traceback.format_exc()))
            
if __name__ == "__main__":
    pass



from gempyp.pyprest.variableReplacement import VariableReplacement
import logging as logger


class MiscVariables:
    def __init__(self, pyprest_obj):
        self.pyprest_obj = pyprest_obj
        logger.info("+++++++++++++++++++++++++++  HANDLING MISC VARIABLES  +++++++++++++++++++++++++++")

  
    def miscVariables(self):
        try:
            if self.pyprest_obj.reporter.__dict__['_misc_data'] is None:
                self.pyprest_obj.reporter.__dict__['_misc_data'] = {}
        except Exception as e:
            print(e)
            logger.error(e)
        try:
            if self.pyprest_obj.report_misc:
                misc_variable_str = self.pyprest_obj.report_misc
                print(misc_variable_str)
                misc_variable_list = misc_variable_str.split(";")
                for each in misc_variable_list:
                    each_item = each.split("=")
                    if "$[#" in each_item[0].strip(" "):
                        key = each_item[0].strip("set $[#").strip("Set $[#").strip("SET $[#").strip("]")
                        if "$[#" in each_item[1]:
                            self.value = VariableReplacement(self.pyprest_obj).getVariableValues(each_item[1].strip(" "))
                            self.pyprest_obj.reporter._misc_data[key.replace("misc.","").upper()] = self.value.upper()
                            logger.info(f"----- ADDED MISC VARIABLE WITH KEY: {key} & VALUE: {self.value} ------")
                        else:
                            self.pyprest_obj.reporter._misc_data[key.replace("misc.","").upper()] = each_item[1].strip(" ").upper()
                            logger.info(f"----- ADDED MISC VARIABLE WITH KEY: {key.strip('misc.').upper()} & VALUE: {each_item[1].strip(' ')} ------")

        except Exception as e:
            logger.error(e)
        logger.info(f"Report_misc_variables:  { self.pyprest_obj.reporter._misc_data } ")

        



from asyncio.log import logger
import os
import traceback
import logging
import json
from typing import Dict, List, Tuple
from gempyp.engine.baseTemplate import TestcaseReporter as Base
from gempyp.libs.enums.status import status
from gempyp.pyprest import apiCommon as api
from gempyp.libs import common
from gempyp.engine.runner import getError
from gempyp.pyprest.reporting import writeToReport
from gempyp.pyprest.preVariables import PreVariables
from gempyp.pyprest.variableReplacement import VariableReplacement as VarReplacement
from gempyp.pyprest.postVariables import PostVariables
from gempyp.pyprest.keyCheck import KeyCheck
from gempyp.pyprest.postAssertion import PostAssertion
from gempyp.pyprest.restObj import RestObj
from gempyp.pyprest.miscVariables import MiscVariables
from gempyp.libs.common import moduleImports
# from gempyp.libs import custom_s3


class PypRest(Base):
    def __init__(self, data) -> Tuple[List, Dict]:
        # self.logger.root.setLevel(self.logger.DEBUG)
        self.data = data
        self.logger = data["config_data"]["LOGGER"] if "LOGGER" in data["config_data"].keys() else logging
        self.logger.info("---------------------Inside REST FRAMEWORK------------------------")
        self.logger.info(f"-------Executing testcase - \"{self.data['config_data']['NAME']}\"---------")
        self.isLegacyPresent = self.isLegacyPresent()

        # set vars
        self.setVars()

        # setting reporter object
        self.reporter = Base(project_name=self.project, testcase_name=self.tcname)
        self.logger.info("--------------------Report object created ------------------------")
        self.reporter.addRow("Starting Test", f'Testcase Name: {self.tcname}', status.INFO) 

    def restEngine(self):
        output = []
        try:
            try:
                self.validateConf()
                self.run()
            except Exception as e:
                if str(e) == "abort":
                    self.logger.info("aborting execution")
                    # self.reporter.addMisc("REASON OF FAILURE","ABORTING EXECUTION")
                else:
                    self.logger.error(str(e))
                    traceback.print_exc()
                    self.logger.error(traceback.print_exc())
                    self.reporter.addMisc("REASON OF FAILURE", f"Something went wrong:- {str(e)}")
                    self.reporter.addRow("Executing Test steps", f'Something went wrong while executing the testcase- {str(e)}', status.ERR)
                    exceptiondata = traceback.format_exc().splitlines()
                    exceptionarray = [exceptiondata[-1]] + exceptiondata[1:-1]
                    self.reporter.addMisc("Reason of Failure",exceptionarray[0])

            VarReplacement(self).valueNotFound()
            output = writeToReport(self)
            return output, None
        except Exception as e:
            self.logger.error(traceback.print_exc())
            common.errorHandler(self.logger, e, "Error occured while running the testcas")
            error_dict = getError(e, self.data["config_data"])
            error_dict["json_data"] = self.reporter.serialize()
            return None, error_dict
    
    def run(self):
        self.getVals()
        # execute and format result
        self.execRequest()
        self.postProcess()
        MiscVariables(self).miscVariables()
        self.logger.info("--------------------Execution Completed ------------------------")
        self.reporter.finalizeReport()

    def validateConf(self):
        mandate = ["API", "METHOD"]
        self.list_subtestcases=[]
        self.request_obj=[]
        self.response_obj=[]
        if len(set(mandate) - set([i.upper() for i in self.data["config_data"].keys()])) > 0:
            # update REASON OF FAILURE in misc
            self.reporter.addMisc("REASON OF FAILURE", "Mandatory keys are missing")
            # self.reporter.addRow("Initiating Test steps", f'Error Occurred- Mandatory keys are missing', status.FAIL)
            raise Exception("mandatory keys missing")
     
        
        if(self.data["config_data"]["RUN_FLAG"]=="Y" and "SUBTESTCASES_DATA" in self.data["config_data"].keys()):
            # self.reporter.addRow("Parent Testcase",f'Testcase Name: {self.data["config_data"]["NAME"]}',status.INFO)
            self.list_subtestcases=self.data["config_data"]["SUBTESTCASES_DATA"]
            
            
            self.variables["local"] = {}
            self.variables["suite"] = self.data["SUITE_VARS"]
            self.parent_data=self.data["config_data"]
            for i in range(len(self.list_subtestcases)):
                self.reporter.addRow("<b>Subtestcase</b>",f'<b>Subtestcase Name: {self.list_subtestcases[i]["NAME"]}</b>',status.INFO)
                self.data["config_data"]=self.list_subtestcases[i]
                self.getVals()

                requestObj=api.Request()
                requestObj.api = self.api
                requestObj.method = self.method
                requestObj.body = self.body
                requestObj.headers = self.headers
                requestObj.file = self.file
                if self.auth_type == "NTLM":
                    requestObj.credentials = {"username": self.username, "password": self.password}
                    requestObj.auth = "PASSWORD"
                self.request_obj.append(requestObj)
              
                self.execRequest()
                
                self.postProcess()
                
                MiscVariables(self).miscVariables()
            del self.parent_data["SUBTESTCASES_DATA"]
            self.data["config_data"]=self.parent_data
            self.request_obj=[]

            
    # read config and get data
    def getVals(self):
        """This is a function to get the values from configData, store it in self object."""
        # capitalize the keys

        for k, v in self.data["config_data"].items():
            self.data.update({k.upper(): v})
        self.env = self.data.get("ENV", "PROD").strip(" ").upper()
        # get the api url
        if self.env not in self.data.keys():
            self.api = self.data["config_data"]["API"].strip(" ")
        else: 
            self.api = self.data.get(self.env, "PROD").strip(" ") + self.data["config_data"]["API"].strip(" ")
        
        # get the method
        self.method = self.data["config_data"].get("METHOD", "GET")

        # get the headers
        # self.headers = self.data["config_data"].get("HEADERS",{})
        self.headers = self.data["config_data"].get("HEADERS", {})
        

        #get miscellaneous variables for report.
        self.report_misc = self.data["config_data"].get("REPORT_MISC","")
        
        # get body
        self.body = self.data["config_data"].get("BODY", {})

        # get file
        self.file = self.data["config_data"].get("REQUEST_FILE", None)

        # get pre variables, not mandatory
        self.pre_variables = self.data["config_data"].get("PRE_VARIABLES", "")

        self.key_check = self.data["config_data"].get("KEY_CHECK", None)

        self.exp_status_code = self.getExpectedStatusCode("EXPECTED_STATUS_CODE") 

        self.post_assertion = self.data["config_data"].get("POST_ASSERTION", None)

        self.post_variables = self.data["config_data"].get("POST_VARIABLES", "")

        self.auth_type = self.data["config_data"].get("AUTHENTICATION", "")

        self.username = self.data["config_data"].get("USERNAME", self.data.get("USER", None))

        self.password = self.data["config_data"].get("PASSWORD", None)

        #get values of mandatory keys of legacy apis
        if self.isLegacyPresent and len(["LEGACY_API", "LEGACY_METHOD", "LEGACY_HEADERS", "LEGACY_BODY"] - self.data["config_data"].keys()) == 0:
            self.legacy_api = self.data["config_data"]["LEGACY_API"].strip(" ")
            self.legacy_method = self.data["config_data"].get("LEGACY_METHOD", "GET")
            self.legacy_headers = json.loads(str(self.data["config_data"].get("LEGACY_HEADERS", {})))
            self.legacy_body = json.loads(str(self.data["config_data"].get("LEGACY_BODY", {})))
            self.legacy_exp_status_code = self.getExpectedStatusCode("LEGACY_EXPECTED_STATUS_CODE")
            self.legacy_auth_type = self.data["config_data"].get("LEGACY_AUTHENTICATION", "")
        #setting variables and variable replacement
        
        PreVariables(self).preVariable()
        VarReplacement(self).variableReplacement()
        self.body=json.loads(self.body) if type(self.body) == str else self.body 
        self.headers=json.loads(self.headers) if type(self.headers) == str else self.headers 

    def file_upload(self,json_form_data): 
        files_data=[]
        json_form_data_1={}  
        for key,value in json_form_data.items():
            files_data_tuple=tuple()
            if(os.path.exists(json_form_data[key])):
                files_data_tuple+=(key,json_form_data[key])
            files_data.append(files_data_tuple)
        return files_data

    def execRequest(self):
        """This function
        -creates a request object, 
        -logs the request
        -runs before method
        -sends request
        -log response 
        -stores it in self object"""
        if(len(self.request_obj)>0):
            self.req_obj=self.request_obj[-1]
            
        else:
            self.req_obj = api.Request()
            # create request
            self.req_obj.api = self.api
            self.req_obj.method = self.method
            self.req_obj.body = self.body
            self.req_obj.headers = self.headers
            if(self.file is not None):
                self.req_obj.file = self.file_upload(json.loads(self.file))
            if self.auth_type == "NTLM":
                self.req_obj.credentials = {"username": self.username, "password": self.password}
                self.req_obj.auth = "PASSWORD"


    
        """legacy api request"""
        # create legacy request
        if hasattr(self,'legacy_api'):
            self.legacy_req = api.Request()
            self.legacy_req.api = self.legacy_api
            self.legacy_req.method = self.legacy_method
            self.legacy_req.headers = self.legacy_headers
            self.legacy_req.body = self.legacy_body  
        # calling the before method after creating the request object.
        self.beforeMethod()
        self.logRequest()
        # calling variable replacement after before method
        
        VarReplacement(self).variableReplacement()
       
        try:
            # raise Exception(f"Error occured while sending request- test")
            self.logger.info("--------------------Executing Request ------------------------")
            self.logger.info(f"url: {self.req_obj.api}")
            self.logger.info(f"method: {self.req_obj.method}")
            self.logger.info(f"request_body: {self.req_obj.body}")
            self.logger.info(f"headers: {self.req_obj.headers}") 

            # addig request misc
            self.reporter.addMisc("REQUEST URL", str(self.req_obj.api)) 
            self.reporter.addMisc("REQUEST METHOD", str(self.req_obj.method))
            self.reporter.addMisc("REQUEST BODY", str(self.req_obj.body))  # s3
            self.reporter.addMisc("REQUEST HEADERS", str(self.req_obj.headers))

            # execute request
            self.res_obj = api.Api().execute(self.req_obj)
            if(len(self.request_obj)>0):
                self.response_obj.append(self.res_obj)
            self.logger.info(f"API response code: {str(self.res_obj.status_code)}")

            # self.reporter.addMisc("RESPONSE BODY", str(self.res_obj.response_body))  # s3
            # self.reporter.addMisc("RESPONSE HEADERS", str(self.res_obj.response_headers))
            self.reporter.addMisc("ACTUAL/EXPECTED RESPONSE CODE", f"{self.res_obj.status_code}/{str(self.exp_status_code).strip('[]')}")

            # logging legacy api
            try:
                # if self.legacy_req is not None:
                self.logger.info("--------------------Executing legacy Request ------------------------")

                self.logger.info(f"legacy url: {self.legacy_req.api}")
                self.logger.info(f"legacy method: {self.legacy_req.method}")
                self.logger.info(f"legacy request_body: {self.legacy_req.body}")
                self.logger.info(f"legacy headers: {self.legacy_req.headers}")

                # addig request misc
                self.reporter.addMisc("LEGACY REQUEST URL", str(self.legacy_req.api))
                self.reporter.addMisc("LEGACY REQUEST METHOD", str(self.legacy_req.method))
                self.reporter.addMisc("LEGACY REQUEST BODY", str(self.legacy_req.body))  # s3
                self.reporter.addMisc("LEGACY REQUEST HEADERS", str(self.legacy_req.headers))

                self.legacy_res = api.Api().execute(self.legacy_req)

                # self.reporter.addMisc("LEGACY RESPONSE BODY", str(self.legacy_res.response_body))  # s3
                # self.reporter.addMisc("LEGACY RESPONSE HEADERS", str(self.legacy_res.response_headers))
                self.reporter.addMisc("ACTUAL/EXPECTED RESPONSE CODE", f"{self.legacy_res.status_code}/{str(self.legacy_exp_status_code).strip('[]')}")


                self.reporter.addMisc("Current Response Time", "{0:.{1}f} sec(s)".format(self.res_obj.response_time,2))
                self.reporter.addMisc("Legacy Response Time", "{0:.{1}f} sec(s)".format(self.legacy_res.response_time,2) )
                self.logResponse()
            except Exception as e:
                if not hasattr(self.legacy_res, "response_body"):
                    self.reporter.addMisc("Response Time", "{0:.{1}f} sec(s)".format(self.res_obj.response_time,2))
                    self.logResponse()
                elif str(e) == "abort":
                    raise Exception("abort")
                if self.isLegacyPresent:
                    traceback.print_exc()
        except Exception as e:
            if str(e) == "abort":
                raise Exception("abort")
            self.logger.info(traceback.print_exc())
            # self.reporter.addRow("Executing API", "Some error occurred while hitting the API", status.FAIL)
            self.reporter.addMisc("REASON OF FAILURE", f"Some error occurred while sending request- {str(e)}")
            raise Exception(f"Error occured while sending request - {str(e)}")

    def setVars(self):
        """
        For setting variables like testcase name, output folder etc.
        """
        self.default_report_path = os.path.join(os.getcwd(), "pyprest_reports")
        self.data["OUTPUT_FOLDER"] = self.data.get("OUTPUT_FOLDER", self.default_report_path)
        if self.data["OUTPUT_FOLDER"].strip(" ") == "":
            self.data["OUTPUT_FOLDER"] = self.default_report_path
        self.project = self.data["PROJECT_NAME"]
        self.tcname = self.data["config_data"]["NAME"]
        self.legacy_req = None
        self.req_obj = None
        self.res_obj = None
        self.legacy_res = None
        self.request_file = None
        self.env = self.data["ENV"]
        self.variables = {}
        self.category = self.data["config_data"].get("CATEGORY", None)
        # self.product_type = self.data["PRODUCT_TYPE"]

    def logRequest(self):
        if self.legacy_req is not None and self.legacy_req.api != '':
            self.logger.info(f"{self.legacy_req.__dict__}")
            self.logger.info(f"{self.req_obj.__dict__}")
            

            legacy_request_body = f"</br><b>REQUEST BODY</b>: {self.legacy_req.body}" if self.legacy_req.method.upper() != "GET" else ""
            current_request_body = f"</br><b>REQUEST BODY</b>: {self.req_obj.body}" if self.req_obj.method.upper() != "GET" else ""
                
            self.reporter.addRow("Executing the rest endpoint","Execution of base api and legacy api simultaneously",
                            status.INFO ,
                            CURRENT_API= f"<b>URL</b>: {self.req_obj.api}</br>" 
                             + f"<b>METHOD</b>: {self.req_obj.method}</br>" 
                             + f"<b>REQUEST HEADERS</b>: {self.req_obj.headers}" 
                             + current_request_body
                            ,LEGACY_API=f"<b>URL</b>: {self.legacy_req.api}</br>" 
                             + f"<b>METHOD</b>: {self.legacy_req.method}</br>" 
                             + f"<b>REQUEST HEADERS</b>: {self.legacy_req.headers}" 
                             + legacy_request_body
            )
        else:
            body_str = f"</br><b>REQUEST BODY</b>: {self.req_obj.body}" if self.req_obj.method.upper() != "GET" else ""
            self.reporter.addRow("Executing the REST Endpoint", 
                             f"<b>URL</b>: {self.req_obj.api}</br>" 
                             + f"<b>METHOD</b>: {self.req_obj.method}</br>" 
                             + f"<b>REQUEST HEADERS</b>: {self.req_obj.headers}" 
                             + body_str, 
                             status.PASS)

    def logResponse(self):
        if self.legacy_res is not None and self.legacy_req.api !='':
            legacy_response_body = self.legacy_res.response_body
            current_response_body = self.res_obj.response_body
            if isinstance(current_response_body, str):
                if "<!DOCTYPE html>" in current_response_body and "</html>" in current_response_body:
                    current_response_body = f"<a href={self.res_obj.api} target = '_blank'>Click here</a>"
            if isinstance(legacy_response_body,str):
                if "<!DOCTYPE html>" in legacy_response_body and "</html>" in legacy_response_body:
                    legacy_response_body = f"<a href={self.legacy_res.api} target = '_blank'>Click here</a>"            
            self.reporter.addRow("Details of the REST endpoint execution","Execution of Current and Legacy API simultaneously",
                            status.INFO ,
                            CURRENT_API= f"<b>CURRENT RESPONSE CODE</b>: {self.res_obj.status_code}</br>" 
                             + f"<b>CURRENT RESPONSE HEADERS</b>: {self.res_obj.response_headers}</br>" 
                             + f"<b>CURRENT RESPONSE BODY</b>:{current_response_body}",
                            LEGACY_API=f"<b>LEGACY STATUS CODE</b>: {self.legacy_res.status_code}</br>" 
                             + f"<b>LEGACY RESPONSE HEADERS</b>: {self.legacy_res.response_headers}</br>" 
                             + f"<b>LEGACY RESPONSE BODY</b>: {legacy_response_body}"
                             )
            if (self.legacy_res.status_code in self.legacy_exp_status_code) and (self.res_obj.status_code in self.exp_status_code) :
                self.reporter.addRow("Validating Response Code with Expected Status Codes", "Both status codes are matching with expected status codes",
                                status.PASS,
                                 CURRENT_API= f"<b>EXPECTED CURRENT RESPONSE CODE</b>: {str(self.exp_status_code).strip('[]')}</br>" 
                                 + f"<b>ACTUAL CURRENT RESPONSE CODE</b>: {str(self.res_obj.status_code)}", 
                                 LEGACY_API= f"<b>EXPECTED LEGACY RESPONSE CODE</b>: {str(self.legacy_exp_status_code).strip('[]')}</br>" 
                                 + f"<b>ACTUAL LEGACY RESPONSE CODE</b>: {str(self.legacy_res.status_code)}"
                                 )
            else:
                self.reporter.addRow("Validating Response Code with Expected Status Codes of both APIs", "Both status codes are not matching with expected status codes",
                                status.FAIL,
                                 CURRENT_API= f"<b>EXPECTED CURRENT RESPONSE CODE</b>: {str(self.exp_status_code).strip('[]')}</br>" 
                                 + f"<b>ACTUAL CURRENT RESPONSE CODE</b>: {str(self.res_obj.status_code)}", 
                                 LEGACY_API= f"<b>EXPECTED LEGACY RESPONSE CODE</b>: {str(self.legacy_exp_status_code).strip('[]')}</br>" 
                                 + f"<b>ACTUAL LEGACY RESPONSE CODE</b>: {str(self.legacy_res.status_code)}"
                                 )
                self.reporter.addMisc("REASON OF FAILURE", "Response code is not as expected")
                self.logger.info("status codes of both apis did not match, aborting testcase.....")
                raise Exception("abort")
                # raise Exception("abort")       
        else:
            body = self.res_obj.response_body
            if isinstance(body, str):
                if "<!DOCTYPE html>" in body and "</html>" in body:
                    body = f"<a href={self.req_obj.api} target = '_blank'>Click here</a>"
        
            self.reporter.addRow("Details of Request Execution", 
                                 f"<b>RESPONSE CODE</b>: {self.res_obj.status_code}</br>" 
                                 + f"<b>RESPONSE HEADERS</b>: {self.res_obj.response_headers}</br>" 
                                 + f"<b>RESPONSE BODY</b>: {str(body)}", 
                                 status.INFO)
            logging.info(f"<b>RESPONSE BODY</b>: {str(body)}")
            if self.res_obj.status_code in self.exp_status_code:
                self.reporter.addRow("Validating Response Code", 
                                 f"<b>EXPECTED RESPONSE CODE</b>: {str(self.exp_status_code).strip('[]')}</br>" 
                                 + f"<b>ACTUAL RESPONSE CODE</b>: {str(self.res_obj.status_code)}", 
                                 status.PASS)
            else:
                self.reporter.addRow("Validating Response Code", 
                                 f"<b>EXPECTED RESPONSE CODE</b>: {str(self.exp_status_code).strip('[]')}</br>" 
                                 + f"<b>ACTUAL RESPONSE CODE</b>: {str(self.res_obj.status_code)}", 
                                 status.FAIL)
                self.reporter.addMisc("REASON OF FAILURE", "Response code is not as expected")
                self.logger.info("status codes did not match, aborting testcase.....")
                raise Exception("abort")

    def postProcess(self):
        """To be run after API request is sent.
        It includes after method, post assertion and post vaiables"""

        PostVariables(self).postVariables()
        KeyCheck(self).keyCheck()
        PostAssertion(self).postAssertion()
        self.afterMethod()

    def beforeMethod(self):
        """This function
        -checks for the before file tag
        -stores package, module,class and method
        -runs before method if found
        -takes all the data from before method and updates the self object"""

        # check for before_file
        self.logger.info("CHECKING FOR BEFORE FILE___________________________")

        file_str = self.data["config_data"].get("BEFORE_FILE", "")
        if file_str == "" or file_str == " ":
            self.logger.info("BEFORE FILE NOT FOUND___________________________")
            self.reporter.addRow("Searching for pre API steps", "No Pre API steps found", status.INFO)

            return
        self.reporter.addRow("Searching for pre API steps", "Searching for before API steps", status.INFO)
        
        file_name = file_str.split("path=")[1].split(",")[0]
        if "CLASS=" in file_str.upper():
            class_name = file_str.split("class=")[1].split(",")[0]
        else:
            class_name = ""
        if "METHOD=" in file_str.upper():
            method_name = file_str.split("method=")[1].split(",")[0]
        else:
            method_name = "before"
        
        self.logger.info("Before file path:- " + file_name)
        self.logger.info("Before file class:- " + class_name)
        self.logger.info("Before file mthod:- " + method_name)
        try:
            # trying to download from s3 path
            # if(file_name.__contains__('s3')):
            #     before_file=file_name.split("/")
            #     folder = before_file[3:]
            #     my_bucket = before_file[2].split(".")[0]
            #     file = before_file[-1]
            #     file_name = custom_s3.download(bucket=my_bucket, file_name=file, folder=folder)
            file_obj = moduleImports(file_name)
            self.logger.info("Running before method")
            obj_ = file_obj
            before_obj = RestObj(
                pg=self.reporter,
                project=self.project,
                request=self.req_obj,
                response=self.res_obj,
                tcname=self.tcname,
                variables=self.variables,
                legacy_req=self.legacy_req,
                legacy_res=self.legacy_res,
                request_file=self.file,
                env=self.env,
            )
            if class_name != "":
                obj_ = getattr(file_obj, class_name)()
            fin_obj = getattr(obj_, method_name)(before_obj)
            self.extractObj(fin_obj)
            
        except Exception as e:
            self.logger.info(traceback.print_exc())
            self.reporter.addRow("Executing Before method", f"Some error occurred while searching for before method- {str(e)}", status.ERR)
        VarReplacement(self).variableReplacement()

    def afterMethod(self):
        """This function
        -checks for the after file tag
        -stores package, module,class and method
        -runs after method if found
        -takes all the data from after method and updates the self object"""

        self.logger.info("CHECKING FOR AFTER FILE___________________________")

        file_str = self.data["config_data"].get("AFTER_FILE", "")
        if file_str == "" or file_str == " ":
            self.logger.info("AFTER FILE NOT FOUND___________________________")
            self.reporter.addRow("Searching for post API steps", "No Post API steps found", status.INFO)
            return

        self.reporter.addRow("Searching for post API steps", "Searching for after method steps", status.INFO)
        
        file_name = file_str.split("path=")[1].split(",")[0]
        if "CLASS=" in file_str.upper():
            class_name = file_str.split("class=")[1].split(",")[0]
        else:
            class_name = ""
        if "METHOD=" in file_str.upper():
            method_name = file_str.split("method=")[1].split(",")[0]
        else:
            method_name = "after"
        self.logger.info("After file path:- " + file_name)
        self.logger.info("After file class:- " + class_name)
        self.logger.info("After file mthod:- " + method_name)
        try:
            # if(file_name.__contains__('s3')):
            #     after_file=file_name.split("/")
            #     folder = after_file[3:]
            #     my_bucket = after_file[2].split(".")[0]
            #     file = after_file[-1]
            #     file_name = custom_s3.download(bucket=my_bucket, file_name=file, folder=folder)
            file_obj = moduleImports(file_name)
            self.logger.info("Running before method")
            obj_ = file_obj
            after_obj = RestObj(
                pg=self.reporter,
                project=self.project,
                request=self.req_obj,
                response=self.res_obj,
                tcname=self.tcname,
                variables=self.variables,
                legacy_req=self.legacy_req,
                legacy_res=self.legacy_res,
                request_file=self.file,
                env=self.env,
            )
            if class_name != "":
                obj_ = getattr(file_obj, class_name)()
            fin_obj = getattr(obj_, method_name)(after_obj)
            self.extractObj(fin_obj)
        except Exception as e:
            self.reporter.addRow("Executing After method", f"Some error occurred while searching for after method- {str(e)}", status.ERR)

        VarReplacement(self).variableReplacement()
        pass

    def extractObj(self, obj):
        """To ofload the data from pyprest obj helper, assign the values back to self object"""

        self.reporter = obj.pg
        self.project = obj.project
        self.req_obj = obj.request
        self.res_obj = obj.response
        self.tcname = obj.tcname
        self.variables = obj.variables
        self.legacy_req = obj.legacy_req
        self.legacy_res = obj.legacy_res
        self.file = obj.request_file
        self.env = obj.env

    def getExpectedStatusCode(self,exp_status_code_param):
        code_list = []
        # self.data["config_data"].get("EXPECTED_STATUS_CODE", 200)
        exp_status_code_string = self.data["config_data"].get(f"{exp_status_code_param}",str(200))
        if "," in exp_status_code_string:
            code_list = exp_status_code_string.strip('"').strip("'").split(",")
        elif "or" in exp_status_code_string.lower():
            code_list = exp_status_code_string.strip('"').strip("'").lower().split("or")
        else:
            code_list = [exp_status_code_string.strip("'").strip(" ").strip('"')]
        code_list = [int(each.strip(" ")) for each in code_list if each not in ["", " "]]
        return code_list
    
    def isLegacyPresent(self):
        if self.data["config_data"].get("LEGACY_API", None) is not None:
            if self.data["config_data"].get("LEGACY_METHOD", None) is not None:
                if json.loads(str(self.data["config_data"].get("LEGACY_HEADERS", {}))) is not None:
                        if json.loads(str(self.data["config_data"].get("LEGACY_BODY", {}))) is not None:
                            if self.getExpectedStatusCode("LEGACY_EXPECTED_STATUS_CODE") is not None:
                                if self.data["config_data"].get("LEGACY_AUTHENTICATION", "") is not None:
                                    return True
                                else:
                                    return False
                            else:
                                return False
                        else:
                            return False
                else:
                    return False
            else:
                return False
        else:
            return False
    


   




from pathlib import Path
import sys
import os
import platform
import getpass
import traceback
from multiprocessing import Pool
from typing import Dict, List, Tuple, Type
from unittest import TestCase
import uuid
from datetime import datetime, timezone
from gempyp.config.baseConfig import AbstarctBaseConfig
from gempyp.engine.testData import TestData
from gempyp.libs.enums.status import status
from gempyp.libs.enums.run_types import RunTypes
from gempyp.reporter.reportGenerator import TemplateData
from gempyp.libs import common
from gempyp.engine.runner import testcaseRunner
from gempyp.config import DefaultSettings
import logging
from gempyp.libs.logConfig import my_custom_logger, LoggingConfig
from gempyp.engine import dataUpload
from gempyp.pyprest.pypRest import PypRest
import smtplib
from gempyp.dv.dvRunner import DvRunner
from gempyp.jira.jiraIntegration import jiraIntegration
from multiprocessing import Process, Pipe
from gempyp.libs.gem_s3_common import upload_to_s3



def executorFactory(data: Dict,conn= None, custom_logger=None ) -> Tuple[List, Dict]:
    
    """
    calls the differnt executors method based on testcase type e.g. gempyp,pyprest,dvm
    Takes single testcase data as input
    """

    print("--------- In Executor Factory ----------\n")
    if custom_logger == None:
        log_path = os.path.join(os.environ.get('TESTCASE_LOG_FOLDER'),data['config_data'].get('NAME') + '_'
        + os.environ.get('unique_id') + '.log')
        custom_logger = my_custom_logger(log_path)
        LoggingConfig(log_path)
    data['config_data']['LOGGER'] = custom_logger
    if 'log_path' not in data['config_data']:
        data['config_data']['LOG_PATH'] = log_path
    

    
    # engine_control = {"pyprest": {"function": PypRest(data).restEngine(), "log": custom_logger.info("Starting the PYPREST testcase")},
    # "dvm": {"function": DvmRunner(data).dvmEngine(), "log": custom_logger.info("Starting the DVM testcase")}, 
    # "gempyp": {"function": testcaseRunner(data), "log": custom_logger.info("Starting the GEMPYP testcase")}}
    engine_control = {
        "pyprest":{"class": PypRest, "classParam": data, "function": "restEngine"},
        "dv":{"class": DvRunner, "classParam": data, "function": "dvEngine"},
        "gempyp":{"function": testcaseRunner, "functionParam": data}
    }

    _type = data.get("config_data").get("TYPE","GEMPYP")
    dv = ["data validator","dv","datavalidator","dvalidator"]
    if _type in dv:
        _type = "dv"

    _type = data.get("config_data")["TYPE"]
    dv = ["data validator","dv","datavalidator","dvvalidator"]
    if _type in dv:
        _type = "dv"

    _type_dict = engine_control[_type.lower()]
    custom_logger.info(f"Starting {_type} testcase")

    if _type_dict.get("class", None):
        data =  getattr(_type_dict["class"](_type_dict['classParam']), _type_dict['function'])()  # need to make it generic for functionParam too
    else:
        data =  _type_dict["function"](_type_dict["functionParam"])  # we need to dissolve this else condition too somehow
    if conn == None:
        return data
    else:
        conn.send([data])
        conn.close()


class Engine:
    def __init__(self, params_config):
        """
        constructor used to  call run method
        takes config as input
        """
        # logging.basicConfig()
        # logging.root.setLevel(logging.DEBUG)
        self.run(params_config)
        

    def run(self, params_config: Type[AbstarctBaseConfig]):
        """
        main method to call other methods that are required for report generation
        takes config as input
        """
        logging.info("Engine Started")
        # initialize the data class
        

        self.DATA = TestData()
        # get the env for the engine Runner
        self.ENV = os.getenv("ENV_BASE", "BETA").upper()
        # initial SETUP
        self.setUP(params_config)
        self.parseMails()
        self.makeSuiteDetails()
        #jewel variable is to print jewel link in rep summary
        self.jewel = ''
        unuploaded_path = ""
        failed_Utestcases = 0
        if self.jewel_user:
            #trying first rerun of base url api in case of api failure
            if self.PARAMS.get("BASE_URL", None) and DefaultSettings.apiSuccess == False:
                logging.info("Retrying to call Api for getting urls")
                DefaultSettings.getEnterPoint(self.PARAMS["BASE_URL"] ,self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"])

            # code for checking s_run_id present in db 
            if "RUN_ID" in self.PARAMS:
                print("************Trying to check If s_run_id is present in DB*****************")
                response =  dataUpload.checkingData(self.s_run_id, self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"])
                if response == "failed":
                    print("************s_run_id not present in DB Trying to call Post*****************")
                    dataUpload.sendSuiteData((self.DATA.toSuiteJson()), self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"])

            else:
                dataUpload.sendSuiteData((self.DATA.toSuiteJson()), self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"])
                ### first try to rerun the data
                if dataUpload.suite_uploaded == False:
                    print("------Retrying to Upload Suite Data------")
                    dataUpload.sendSuiteData((self.DATA.toSuiteJson()), self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"])
            
        self.makeOutputFolder()
        self.start()

        if(self.jewel_user):
            #trying second rerun of base url api in case of api failure
            if self.PARAMS.get("BASE_URL", None) and DefaultSettings.apiSuccess == False:
                logging.info("Second Time Retrying to call Api for getting urls")
                DefaultSettings.getEnterPoint(self.PARAMS["BASE_URL"] ,self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"] )
            ### Trying to reupload suite data
            if dataUpload.suite_uploaded == False:
                print("------Retrying to Upload Suite Data------")
                dataUpload.sendSuiteData((self.DATA.toSuiteJson()), self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"])

        ### checking if suite data is uploaded if true than retrying to upload testcase otherwise storing them in json file
        if dataUpload.suite_uploaded == True:
            jewelLink = DefaultSettings.getUrls('jewel-url')
            self.jewel = f'{jewelLink}/#/autolytics/execution-report?s_run_id={self.s_run_id}'
            if len(dataUpload.not_uploaded) != 0:
                print("------Trying again to Upload Testcase------")
                for testcase in dataUpload.not_uploaded:
                    dataUpload.sendTestcaseData(testcase, self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"])
            failed_Utestcases = len(dataUpload.not_uploaded) 
            ### Creating file for unuploaded testcases
            if len(dataUpload.not_uploaded) != 0:
                if dataUpload.flag == True:
                    logging.warning("Testcase may be present with same tc_run_id in database")
                listToStr = ',\n'.join(map(str, dataUpload.not_uploaded))
                unuploaded_path = os.path.join(self.ouput_folder, "Unploaded_testCases.json")
                with open(unuploaded_path,'w') as w:
                    w.write(listToStr)
        self.updateSuiteData()
        # suite_status = self.DATA.suite_detail.to_dict(orient="records")[0]["status"]
        # testcase_info = self.DATA.suite_detail.to_dict(orient="records")[0]["testcase_info"]
        skip_jira = 0
        try:
            jira_email = self.PARAMS.get("JIRA_EMAIL", None)
            jira_access_token = self.PARAMS.get("JIRA_ACCESS_TOKEN", None)
            jira_project_id = self.PARAMS.get("JIRA_PROJECT_ID", None)
            jira_workflow = self.PARAMS.get("JIRA_WORKFLOW", None)
            if jira_access_token is None and jira_email is None:
                skip_jira = 1
        except Exception as e:
            pass

        ### checking if suite post/get request is successful to call put request otherwise writing suite data in a file
        if dataUpload.suite_uploaded == True:
            dataUpload.sendSuiteData(self.DATA.toSuiteJson(), self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"], mode="PUT")

            if skip_jira == 0:
                jira_id = jiraIntegration(self.s_run_id, jira_email, jira_access_token, jira_project_id, self.ENV, jira_workflow, self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"], self.report_name)
                if jira_id is not None:
                    self.DATA.suite_detail.at[0, "meta_data"].append({"Jira_id": jira_id})
            # dataUpload.sendSuiteData(self.DATA.toSuiteJson(), self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"], mode="PUT")
        else:
            if not self.PARAMS.get("BASE_URL", None):
                logging.warning("Maybe username or bridgetoken is missing or wrong thus data is not uploaded in db.")
            dataUpload.suite_data.append(self.DATA.toSuiteJson())
            listToStr = ',\n'.join(map(str, dataUpload.suite_data))
            unuploaded_path = os.path.join(self.ouput_folder, "Unuploaded_suiteData.json")
            with open(unuploaded_path,'w') as w:
                w.write(listToStr)
        self.repJson, output_file_path = TemplateData().makeSuiteReport(self.DATA.getJSONData(), self.testcase_data, self.ouput_folder)
        TemplateData().repSummary(self.repJson, output_file_path, self.jewel, failed_Utestcases, unuploaded_path)

    def makeOutputFolder(self):
        """
        make outputFolder for report named as gempyp_reports in user home directory if not given by the user and makes log fplder for log files
        if given by user than set user given path for reports file   
        """

        logging.info("---------- Making output folders -------------")
        report_folder_name = f"{self.project_name}_{self.project_env}"
        if self.report_name:
            report_folder_name = report_folder_name + f"_{self.report_name}"
        date = datetime.now().strftime("%Y_%b_%d_%H%M%S_%f")
        report_folder_name = report_folder_name + f"_{date}"
        if "OUTPUT_FOLDER" in self.PARAMS and self.PARAMS["OUTPUT_FOLDER"]:
            self.ouput_folder = os.path.join(
                self.PARAMS["OUTPUT_FOLDER"], report_folder_name
            )
        else:
            home = str(Path.home())
            self.ouput_folder = os.path.join(
                home, "gempyp_reports", report_folder_name
            )

        os.makedirs(self.ouput_folder)
        self.testcase_folder = os.path.join(self.ouput_folder, "testcases")
        os.makedirs(self.testcase_folder)
        self.testcase_log_folder = os.path.join(self.ouput_folder, "logs")
        os.environ['TESTCASE_LOG_FOLDER'] = self.testcase_log_folder
        os.makedirs(self.testcase_log_folder)

    def setUP(self, config: Type[AbstarctBaseConfig]):
        """
        assigning values to some attributes which will be used in method makeSuiteDetails
        """
        self.PARAMS = config.getSuiteConfig()
        #checking if url is present in file and calling get api
        # if self.PARAMS.get("BASE_URL", None):
        #     DefaultSettings.getEnterPoint(self.PARAMS["BASE_URL"] ,self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"] )
        self.CONFIG = config
        self.testcase_data = {}
        self.total_runable_testcase = config.total_yflag_testcase
        self.machine = platform.node()
        self.user = self.PARAMS.get("USERNAME", getpass.getuser())
        self.invoke_user = os.getenv("INVOKEUSER", self.user)  # INVOKEUSER can be set as environment variable from anywhere.
        self.current_dir = os.getcwd()
        self.platform = platform.system()
        self.start_time = datetime.now(timezone.utc)
        self.project_name = self.PARAMS["PROJECT"]
        self.report_name = self.PARAMS.get("REPORT_NAME")
        self.project_env = self.PARAMS["ENV"]
        self.unique_id = self.PARAMS["UNIQUE_ID"]
        self.user_suite_variables = self.PARAMS["SUITE_VARS"]
        self.jewel_run = False
        self.jewel_user = False
        self.s3_url = ""
        if self.PARAMS.get("BRIDGE_TOKEN", None) and self.PARAMS.get("USERNAME", None):
            self.jewel_user = True
        if self.jewel_user:
            if self.PARAMS.get("BASE_URL", None):
                DefaultSettings.getEnterPoint(self.PARAMS["BASE_URL"] ,self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"])
            if self.PARAMS.get("S_ID", None):
                self.jewel_run = True
            else:
                try:
                    self.s3_url = upload_to_s3(DefaultSettings.urls["data"]["bucket-file-upload-api"], bridge_token=self.PARAMS["BRIDGE_TOKEN"], username=self.PARAMS["USERNAME"], file=self.PARAMS["config"])[0]["Url"]
                    print("--------- url", self.s3_url)
                except Exception as e:
                    print(e)
        #add suite_vars here 

    def parseMails(self):
        """
        to get the mail from the configData
        """
        if("MAIL" in self.PARAMS.keys()):
            self.mail = common.parseMails(self.PARAMS["MAIL"])
            print(self.mail)

    def makeSuiteDetails(self):
        """
        making suiteDetails dictionary and assign it to DATA.suiteDetail 
        """
        if "RUN_ID" in self.PARAMS:
            self.s_run_id = self.PARAMS["RUN_ID"]
        else:
            if not self.unique_id:
                self.unique_id = uuid.uuid4()
            self.s_run_id = f"{self.project_name}_{self.project_env}_{self.unique_id}"
            self.s_run_id = self.s_run_id.upper()
        logging.info("S_RUN_ID: {}".format(self.s_run_id))
        suite_details = {
            "s_run_id": self.s_run_id,
            "s_start_time": self.start_time,
            "s_end_time": None,
            "s_id": self.PARAMS.get("S_ID", "test_id"),
            "status": status.EXE.name,
            "project_name": self.project_name,
            "report_name": self.report_name,  # earlier it was report info
            "user": self.user,
            "env": self.project_env,
            "machine": self.machine,
            "os": platform.system().upper(),
            "meta_data": [],
            "expected_testcases": self.total_runable_testcase,
            "testcase_info": None,
            "framework_name": "GEMPYP",
        }
        self.DATA.suite_detail = self.DATA.suite_detail.append(
            suite_details, ignore_index=True
        )

    def start(self):

        """
         check the mode and start the testcases accordingly e.g.optimize,parallel
        """
        print("here")
        try:
            if self.CONFIG.getTestcaseLength() <= 0:
                raise Exception("no testcase found to run")

            if self.PARAMS["MODE"].upper() == "SEQUENCE":
                self.startSequence()
            elif self.PARAMS["MODE"].upper() == "OPTIMIZE" or self.PARAMS.get("MODE", None) is None:
                self.startParallel()
            else:
                raise TypeError("mode can only be sequence or optimize")

        except Exception as e:
            logging.error(traceback.format_exc())
            try:
                self.DATA.suite_detail.at[0, "meta_data"].append({"REASON OF FAILURE": str(e)})
                self.updateSuiteData()
            except Exception as err:
                logging.error(traceback.format_exc())
                print(err)
            dataUpload.sendSuiteData((self.DATA.toSuiteJson()), self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"])
            # need to add reason of failure of the suite in misc


    def updateSuiteData(self):
        """
        updates the suiteData after all the runs have been executed
        """

        # get the status count of the status
        status_dict = self.DATA.testcase_details["status"].value_counts().to_dict()
        total = sum(status_dict.values())
        status_dict["TOTAL"] = total
        unsorted_dict = status_dict
        sorted_dict = self.totalOrder(unsorted_dict)
        status_dict = sorted_dict
        # status_dict = dict( sorted(status_dict.items(), key=lambda x: x[0].lower(), reverse=True) )
        Suite_status = status.FAIL.name

        # based on the status priority
        for s in status:
            if status_dict.get(s.name, 0) > 0:
                Suite_status = s.name
                break
        stop_time = (
            self.DATA.testcase_details["end_time"].sort_values(ascending=False).iloc[0]
        )
        self.DATA.suite_detail.at[0, "status"] = Suite_status
        self.DATA.suite_detail.at[0, "s_end_time"] = stop_time
        self.DATA.suite_detail.at[0, "testcase_info"] = status_dict
        if self.jewel_run is True:
            self.DATA.suite_detail.at[0, "meta_data"].append({"S_ID": self.PARAMS["S_ID"]})
        else:
            self.DATA.suite_detail.at[0, "meta_data"].append({"CONFIG_S3_URL": self.s3_url})

    def startSequence(self):
        """
        start calling executoryFactory() for each testcase one by one according to their dependency
        at last of each testcase calls the update_df() 
        """

        for testcases in self.getDependency(self.CONFIG.getTestcaseConfig()):
            for testcase in testcases:
                data = self.getTestcaseData(testcase['NAME'])
                log_path = os.path.join(self.testcase_log_folder,
                data['config_data'].get('NAME')+'_'+self.CONFIG.getSuiteConfig()['UNIQUE_ID'] + '.log')
                custom_logger = my_custom_logger(log_path)
                data['config_data']['log_path'] = log_path
                conn = None
                output, error = executorFactory(data,conn, custom_logger)
                if error:
                    custom_logger.error(
                        f"Error occured while executing the testcase: {error['testcase']}"
                    )
                    custom_logger.error(f"message: {error['message']}")
                self.update_df(output, error)



    def startParallel(self):
        """
        start calling executorFactory for testcases in parallel according to their drependency 
        at last of each testcase calls the update_df()
        """
        
        pool = None
        try:
            
            threads = self.PARAMS.get("THREADS", DefaultSettings.THREADS)
            try:
                threads = int(threads)
            except:
                threads = DefaultSettings.THREADS
            # pool = Pool(threads)
           
           
            for testcases in self.getDependency(self.CONFIG.getTestcaseConfig()):

        # create a list to keep connections
                if len(testcases) == 0:
                    raise Exception("No testcase to run")
                pool_list = []
                for testcase in testcases:
                    # only append testcases whose dependency are passed otherwise just update the databasee
                    if self.isDependencyPassed(testcase):
                        pool_list.append(self.getTestcaseData(testcase.get("NAME")))
                    
                    else:

                        print("----------------here--------------------")
                        dependency_error = {
                            "message": "dependency failed",
                            "testcase": testcase["NAME"],
                            "category": testcase.get("CATEGORY", None),
                            "product_type": testcase.get("product_type", None),
                        }
                        # handle dependency error in json_data(update_df)
                        # update the testcase in the database with failed dependency
                        self.update_df(None, dependency_error)
       
                if len(pool_list) == 0:
                    continue
                
                # obj = Semaphore(8)
                splitedSize = threads
                a_splited = [pool_list[x:x+splitedSize] for x in range(0, len(pool_list), splitedSize)]
                for chunk_list in a_splited:
                    processes = []
                    parent_connections = []
                    for testcase in chunk_list:
                        parent_conn, child_conn = Pipe()
                        parent_connections.append(parent_conn)

                # create the process, pass instance and connection
                        custom = None
                        process = Process(target=executorFactory, args=(testcase, child_conn, custom))
                        processes.append(process) 
                # runs the testcase in parallel here
                    instances_total = []
                
                    # print(a_splited)
                    # for i in a_splited:
                    for process in processes:
                        process.start()
                    for parent_connection in parent_connections:
                        instances_total.append(parent_connection.recv()[0])
                        
                    for row in instances_total:
                        if not row or len(row) < 2:
                            raise Exception(
                                "Some error occured while running the testcases"
                            )
                        output = row[0]
                        
                        error = row[1]
                        if error:
                            logging.error(
                                f"Error occured while executing the testcase: {error['testcase']}"
                            )
                            logging.error(f"message: {error['message']}")
                        self.update_df(output, error)
                    for process in processes:
                        process.join()
        except Exception:
            logging.error(traceback.format_exc())


    def update_df(self, output: List, error: Dict):

        """
        updates the testcase data in the dataframes of testData.py
        also upload testcasedata to db
        
        """
        try:
            if error:
                output = self.getErrorTestcase(
                    error["message"],
                    error["testcase"],
                    error.get("category"),
                    error.get("product_type"),
                    error.get('log_path', None),
                    error.get('invoke_user', None),
                )
                output = [output]
            unsorted_dict = output[0]['json_data']['meta_data'][2]
            sorted_dict = self.totalOrder(unsorted_dict)
            output[0]['json_data']['meta_data'][2] = sorted_dict

            output[0]['testcase_dict']['run_type'], output[0]['testcase_dict']['run_mode'] = self.set_run_type_mode()
            for i in output:

                i["testcase_dict"]["steps"] = i["json_data"]["steps"]
                
                testcase_dict = i["testcase_dict"]
                try:
                    """ update suite vars here from testcase_dict["suite_variables"] append it in the suite vars of _config"""
    
                    self.user_suite_variables.update(i.get("suite_variables", {}))
                    
                    self.testcase_data[testcase_dict.get("tc_run_id")] = i["json_data"]
                except Exception as e:
                    logging.error(e)
                self.DATA.testcase_details = self.DATA.testcase_details.append(
                    testcase_dict, ignore_index=True
                )
                self.updateTestcaseMiscData(
                    i["misc"], tc_run_id=testcase_dict.get("tc_run_id")
                )
                if(self.jewel_user):
                    dataUpload.sendTestcaseData((self.DATA.totestcaseJson(testcase_dict.get("tc_run_id").upper(), self.s_run_id)), self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"])
        except Exception as e:
            traceback.print_exc()
            logging.error("in update_df: {e}".format(e=e))

    def set_run_type_mode(self):
        type, mode = None, None
        try:
            if self.PARAMS.get('RUN_TYPE', RunTypes.ON_DEMAND.value) in [i.value for i in RunTypes]:
                type = self.PARAMS.get('RUN_TYPE', RunTypes.ON_DEMAND.value)

            operating_system = "cli-" + platform.uname().system
            mode = self.PARAMS.get('RUN_MODE', operating_system.upper())
        except Exception as e:
            traceback.print_exc()
        return type, mode

    def getErrorTestcase(
        self,
        message: str,
        testcase_name: str,
        category: str = None,
        product_type: str = None,
        log_path: str = None,
        invoke_user: str = None
    ) -> Dict:
        """
        store the data of failed testcase and return it as a dict to update_df
        take message for error as input
        """

        result = {}
        testcase_dict = {}
        misc = {}
        if not self.unique_id:
            self.unique_id = uuid.uuid4()
        tc_run_id = f"{testcase_name}_{self.unique_id}"  # testcase should not be testcase + s_run_id
        # try:
        #     unique_id = unique_id.split(self.env.upper()).strip("_")
        # except Exception as e:
        #     print(e)
        # tc_run_id = f"{testcase_name}_{self.project_env}_{uuid.uuid4()}"
        tc_run_id = tc_run_id.upper()
        testcase_dict["tc_run_id"] = tc_run_id
        testcase_dict["status"] = status.FAIL.name
        testcase_dict["start_time"] = datetime.now(timezone.utc)
        testcase_dict["end_time"] = datetime.now(timezone.utc)
        testcase_dict["name"] = testcase_name
        testcase_dict["ignore"] = False
        if category:
            testcase_dict["category"] = category
        testcase_dict["log_file"] = log_path
        testcase_dict["result_file"] = None
        testcase_dict["base_user"] = getpass.getuser()
        testcase_dict["invoke_user"] = self.invoke_user
        testcase_dict["machine"] = self.machine
        # testcase_dict["response_time"]="{0:.{1}f} sec(s)".format((testcase_dict["end_time"]-testcase_dict["start_time"]).total_seconds(),2)
        if product_type:
            testcase_dict["product_type"] = product_type
        result["testcase_dict"] = testcase_dict
        misc["REASON OF FAILURE"] = message
        result["misc"] = misc
        return result

    def updateTestcaseMiscData(self, misc: Dict, tc_run_id: str):
        """
        updates the misc data for the testcases in testData.py
        accept miscellaneous rows and tc_run_id as parameters
        """
        misc_list = []

        for misc_data in misc:
            temp = {}
            # storing all the key in upper so that no duplicate data is stored
            temp["key"] = misc_data.upper()
            temp["value"] = misc[misc_data]
            temp["run_id"] = tc_run_id
            temp["table_type"] = "TESTCASE"
            misc_list.append(temp)

        self.DATA.misc_details = self.DATA.misc_details.append(
            misc_list, ignore_index=True
        )

    def getTestcaseData(self, testcase: str) -> Dict:
        """
        taking argument as the testcase name and  return dictionary containing information about testCase
        """
        data = {}
        list_subtestcases=[]
        data["config_data"] = self.CONFIG.getTestcaseData(testcase)
        if("SUBTESTCASES" in data["config_data"].keys()):
            for key1 in data["config_data"]["SUBTESTCASES"].split(","):
                list_subtestcases.append(self.CONFIG.getSubTestcaseData(key1))
            data["config_data"]["SUBTESTCASES_DATA"]=list_subtestcases
        data["PROJECT_NAME"] = self.project_name
        data["ENV"] = self.project_env
        if(self.project_env.upper() in self.PARAMS.keys()):
            data[self.project_env.upper()]=self.PARAMS[self.project_env.upper()]
        data["S_RUN_ID"] = self.s_run_id
        # data["USER"] = self.user
        data["MACHINE"] = self.machine
        data["OUTPUT_FOLDER"] = self.testcase_folder
        data["SUITE_VARS"] = self.user_suite_variables
        data["INVOKE_USER"] = self.invoke_user
        data["USER"] = self.user
        return data

    

    def getDependency(self, testcases: Dict):
        """
        yields the testcases with least dependncy first
        Reverse toplogical sort
        accept all testcases dictionary as arguments
        """
        adj_list={}
        for key,value in testcases.items():
                adj_list[key]=list(set(list(value.get("DEPENDENCY", "").upper().split(",")))  - set([""]))       

        for key, value in adj_list.items():
            new_list = []
            for testcase in value:

                testcase = testcase.split(":")
                if len(testcase) > 1:
                    new_list.append(testcase[1])
                else:
                    new_list.append(testcase[0])

            adj_list[key] = set(new_list)
        
        while adj_list:
            """
            return testcases that doesn't depend on other testcase
            """
            top_dep = set(
                i for dependents in list(adj_list.values()) for i in dependents
            ) - set(adj_list.keys())
            top_dep.update(key for key, value in adj_list.items() if not value)
        
            if not top_dep:
                logging.critical(
                    "circular dependency found please remove the cirular dependency"
                )
                logging.debug("possible testcase with circular dependencies")
                sys.exit(1)

            adj_list = {key: value - top_dep for key, value in adj_list.items() if value}
            result = []
            for key in testcases:
                if key in top_dep:
                    result.append(testcases[key])
            yield result

   

    def isDependencyPassed(self, testcase: Dict) -> bool:
        """
        cheks if the dependency is passed for the testcase or not
        """

        # split on ','
        listOfTestcases=[]
        listOfTestcases=list(set(list(testcase.get("DEPENDENCY", "").upper().split(","))) - set([""]))
        for dep in listOfTestcases:

                    dep_split = list(dep.split(":"))
                    if len(dep_split) == 1:
                        # NAME to name, to_list()
                        if dep_split[0] not in self.DATA.testcase_details["name"].to_list():
                            return False

                    else:
                        if dep_split[0].upper() == "P":
                            if dep_split[1] not in self.DATA.testcase_details["name"].to_list():
                                return False
                            # way to parsing the df    
                            if ((self.DATA.testcase_details[self.DATA.testcase_details["name"] == dep_split[1]]['status'].iloc[0]) != status.PASS.name):
                                return False

                        if dep_split[0].upper() == "F":
                            if dep_split[1] not in self.DATA.testcase_details["name"].to_list():
                                return False
                            if (
                                (self.DATA.testcase_details[self.DATA.testcase_details["name"] == dep_split[1]]['status'].iloc[0])
                                != status.FAIL.name
                            ):
                                return False

        return True

    ### Function for sending total, pass, fail in order
    def totalOrder(self, unsorted_dict):
        
        prio_list = ['TOTAL', 'PASS', 'FAIL']
        sorted_dict = {}
        for key in prio_list:
            if key in unsorted_dict :
                sorted_dict[key] = unsorted_dict[key]
                unsorted_dict.pop(key)
            elif key == "PASS" or key == 'FAIL':
                sorted_dict[key] = 0
        sorted_dict.update(unsorted_dict)

        return sorted_dict



from gempyp.config.xmlConfig import XmlConfig
import argparse
from gempyp.engine.engine import Engine
from gempyp.config.GitLinkXML import fetchBridgeToken

class Gempyp:
    def __init__(self):
        """
        initiating some variables  
        """
        self.config = None
        self.MAIL = None
        self.PROJECT = None
        self.REPORT_NAME = None
        self.MODE = None
        self.ENV = None
        self.args = None
        self.THREADS = None
        self.BRIDGE_TOKEN = None 
        self.OUTPUT_FOLDER = None
        self.CATEGORY=None
        self.USERNAME=None
        self.RUN_ID = None
        self.TESTCASE_LIST = None
        self.BASE_URL = None
        self.S_ID = None
        self.RUN_TYPE = None
        self.RUN_MODE = None
    
    def argParser(self):
        """Argument parser to help running through CLI"""

        parser = argparse.ArgumentParser()
        parser.add_argument('-config','-c',dest='config',type=str, required=False)
        parser.add_argument('-mail','-m',dest='MAIL', type=str, required=False)
        parser.add_argument('-project','-p',dest='PROJECT', type=str, required=False)
        parser.add_argument('-Report_name','-rn',dest='REPORT_NAME', type=str, required=False)
        parser.add_argument('-mode','-mode',dest='MODE', type=str, required=False)
        parser.add_argument('-env','-env',dest='ENV', type=str, required=False)
        parser.add_argument('-threads','-t',dest='THREADS',type=str, required=False)
        parser.add_argument('-bridge_token','-token',dest='BRIDGE_TOKEN',type=str, required=False)
        parser.add_argument('-username','-username',dest='USERNAME',type=str, required=False)
        parser.add_argument('-output_folder','-of',dest='OUTPUT_FOLDER',type=str, required=False)
        parser.add_argument('-category','-category',dest='CATEGORY',type=str, required=False)
        parser.add_argument('-run_id','-run_id',dest='RUN_ID',type=str, required=False)
        parser.add_argument('-tc','-testcase_list',dest='TESTCASE_LIST',type=str, required=False)
        parser.add_argument('-base_url','-base_url',dest='BASE_URL',type=str, required=False)
        parser.add_argument('-s_id','-s_id',dest='S_ID',type=str, required=False)
        parser.add_argument('-run_type','-run_type',dest='RUN_TYPE',type=str, required=False)
        parser.add_argument('-run_mode','-run_mode',dest='RUN_MODE',type=str, required=False)

        args = parser.parse_args()
        return args

    def runner(self):
        """
        This function takes the config and updates the config data in case or cli run and direct(python) run
        """
        s_run_id = vars(self)["RUN_ID"]
        if("GIT" in self.config):
            list_url=self.config.split(":")
            config=XmlConfig(fetchBridgeToken(list_url[2],list_url[3],list_url[4],list_url[5]),s_run_id)
        else:
            config = XmlConfig(self.config, s_run_id)
        if not self.args:
            del self.__dict__["args"]
            config.cli_config = vars(self)
        else:
            config.cli_config = vars(self.args)
        config.update()
        Engine(config)

    def parser(self):
        """Calls the parser and handles the case of no cli args"""
        args = self.argParser()
        if args.config != None:
            self.config = args.config
        self.args = args
        self.runner()

def main():
    obj = Gempyp()
    #obj.config = "C:\\Users\\an.pandey\\gempyp\\tests\\configTest\\Gempyp_Test_Suite.xml"
    #obj.ENV = ""
    #obj.MAIL = ""
    obj.parser()

if __name__ == "__main__":
    obj = Gempyp()
    
    #obj.config = "C:\\Users\\an.pandey\\gempyp\\tests\\configTest\\Gempyp_Test_Suite.xml"
    #obj.ENV = ""
    #obj.MAIL = ""
    obj.parser()
    




class Validator():
    """
    This class will validate the config data provided by user. Minimum/maximum values, input data type etc
    """
    def __init__(self, config_data) -> None:
        self.config_data = config_data
    
    def validate_conf(self):
        """ To be done later"""
        return True


import datetime
import os
import sys
import traceback
import threading
import queue
import data_compare.data.database as database

def get_data_thread(config_data, queue1, db_obj = None):
    try:
        print("Data Thread started")
        query = config_data["query"]
        #base_query = config_data["base_query"]
        if not db_obj:
            #will update the db type getting when we will have more type of database
            #add for sqlite
            #db_type = 'sqlite'
            db_type=config_data['type']
            d_obj = database.Database(db_type, config_data)
            db_obj = d_obj.get_db_obj()
            db_obj.connect()
            print("Connected successfully to db")
        #r_val = db_obj.execute_dm_query(query)
        data, header = db_obj.execute_query(query, header=True)
        print("Query executed")
        if not data:
            print("Error in fetching data for query {}".format(query))
        db_obj.close()
        queue1.put((data, header))
        print("Data thread finished")
    except Exception:
        print("Exception in get_data_thread {}".format(traceback.format_exc()))

    finally:
        try:
            db_obj.close()
        except Exception:
            pass

class Data():

    def __init__(self, config_data) -> None:
        self.config_data = config_data
        pass

    def get_data(self):
        try:
            if(self.config_data["src_type"]=="sqlite"):
                src_config = {
                    "type":self.config_data["src_type"],
                    "location":self.config_data["src_location"],
                    "table":self.config_data["srcTable"],
                    #"base_query": "select * from "+self.config_data["srcTable"],
                    "query":self.config_data["src_query_path"],
                    "schema": self.config_data["srcSchema"]
                }
            else:
                src_config = {
                    "type":self.config_data["src_type"],
                    "table":self.config_data["srcTable"],
                    "database":self.config_data["src_database"],
                    #"base_query": "select * from "+self.config_data["srcTable"],
                    "query":self.config_data["src_query_path"],
                    "schema": self.config_data["srcSchema"],
                    "user": self.config_data["src_user"],
                    "password": self.config_data["src_password"],
                    "port":self.config_data["src_port"],
                    "host":self.config_data["src_host"]
                }
            if(self.config_data["tgt_type"]=="sqlite"):	
                tgt_config = {
                    "type":self.config_data["tgt_type"],
                    "location":self.config_data["tgt_location"],
                    "table":self.config_data["tgtTable"],
                    #"base_query": "select * from "+self.config_data["tgtTable"],
                    "query":self.config_data["tgt_query_path"],
                    "schema": self.config_data["tgtSchema"]
                }
            else:
                tgt_config = {
                    "type":self.config_data["tgt_type"],
                    "table":self.config_data["tgtTable"],
                    "database":self.config_data["tgt_database"],
                    #"base_query": "select * from "+self.config_data["tgtTable"],
                    "query":self.config_data["tgt_query_path"],
                    "schema": self.config_data["tgtSchema"],
                    "user": self.config_data["tgt_user"],
                    "password": self.config_data["tgt_password"],
                    "port":self.config_data["tgt_port"],
                    "host":self.config_data["tgt_host"]
                }
            
            if os.path.isfile(src_config["query"]) :
                with open(src_config["query"]) as fobj:
                    print("srcquery file is {}".format(src_config["query"]))
                    src_config["query"] = " ".join(fobj.readlines())
            if os.path.isfile(tgt_config["query"]) :
                with open(tgt_config["query"]) as fobj:
                    print("tgtquery file is {}".format(tgt_config["query"]))
                    tgt_config["query"] = " ".join(fobj.readlines())
            src_queue = queue.Queue()
            tgt_queue = queue.Queue()
            src_thread = threading.Thread(target= get_data_thread, args=(src_config, src_queue))
            src_thread.start()
            print("Src thread started")
            tgt_thread = threading.Thread(target=get_data_thread,args=(tgt_config,tgt_queue))
            tgt_thread.start()
            print("target thread started")
            src_thread.join()
            tgt_thread.join()
            src_data, src_header, tgt_data, tgt_header = None, None, None, None
            if not src_queue.empty():
                src_data,src_header = src_queue.get()
            if not tgt_queue.empty():
                tgt_data,tgt_header = tgt_queue.get()
            #print('1',src_data) 
            #print('2',src_header)
            #print('3',tgt_data)
            #print('4',tgt_header)
            return (src_data,src_header,tgt_data,tgt_header)
            pass
        except Exception:
            print("Exception in get_data {}".format(traceback.format_exc()))
            return (None, None, None, None)



import os
import sys
from libs import common 
from data_compare import compare 
import traceback
import argparse

def test():
   print("hello1")

def test2():
   print("hello3")
   return "hello3"

def parse_argument():
    parser = argparse.ArgumentParser()
    parser.add_argument("-s","--suite",help="suite Config File Path")
    parser.add_argument("-tc","--testcase",help="test Config File Path")

    args = parser.parse_args()
    return args

class GemPyp():
     def __init__(self, s_config_data, t_config_data):
         #read the conf here
         #create suite level result dictionary here
         #create test level result dictionary here
         self.suite_conf = s_config_data  # not using this data as of now.
         self.test_conf = t_config_data
         # adding only with execute =Yes 
         self.test_conf = {x : self.test_conf[x] for x in self.test_conf if self.test_conf[x]["execute"].lower() == 'yes'}
         
     
     def execute(self):
         #create a pool of process here
         # create testcase object depending upon type of testcase and call the function run using that object
         
         #update the result for each test case
         #identifying independent task and tasks which have dependency here, create task roster here for executor to run
         
         #running sequentially, update this code with engine logic
         try:
             for test in self.test_conf:
                 test_type = self.test_conf[test]["test_type"].lower()
                 if test_type == "data_comp": 
                     obj = compare.Compare(test, self.test_conf[test])
                     obj.run()
                 else:
                     # dynamicaly load rclass and execute the run function
                     pass
         except Exception: 
            traceback.print_exc()  
     def create_report(self):
         pass

if __name__ == "__main__":
    args = parse_argument()
    if args:
        conf_dir = os.getenv('CONF_DIR')
        s_conf_file_path = os.path.join(conf_dir, args.suite)
        t_conf_file_path = os.path.join(conf_dir, args.testcase)
        if os.path.isfile(s_conf_file_path) and os.path.isfile(t_conf_file_path):
            s_config_data = common.read_json(s_conf_file_path)
            t_config_data = common.read_json(t_conf_file_path)
        else:
            print(f"config file{s_conf_file_path} does not exist")
            sys.exit(1)
        obj = GemPyp(s_config_data, t_config_data)
        obj.execute()
    else:
         print("Argument parsing had issues")



import os
from gempyp.engine import dataUpload
import traceback
import logging

# have to discuss about the default location
DEFAULT_GEMPYP_FOLDER = os.getcwd()
DEBUG = True
THREADS = 8
_VERSION = "1.0.0"
apiSuccess = False
urls = {"data":{
                "jewel-url": "https://jewel.gemecosystem.com",
                "suiteInfo": "https://apis.gemecosystem.com/suiteinfo/",
                "suite-exe-api": "https://apis.gemecosystem.com/suiteexe",
                "test-exe-api": "https://apis.gemecosystem.com/testcase",
                "last-five": "https://apis.gemecosystem.com/suiteexe/lastFive",
                "jira-api": "https://apis.gemecosystem.com/jira",
                "bucket-file-upload-api": "https://apis-beta.gemecosystem.com/v1/upload/file"
            }
        }
# for getting urls using url tag from config file
def getEnterPoint(url, bridge_token, user_name):
    global urls
    try:
        url = checkUrl(url)
        response = dataUpload._sendData(" ", url, bridge_token, user_name,"GET")
        if response.status_code == 200:
            url_enter_point = response.json()
            urls["data"].update(url_enter_point["data"])
            global apiSuccess
            apiSuccess = True
        else:
            logging.warning("Error Occurs While Getting the BASE_URLs")
    except Exception as e:
            traceback.print_exc()
            logging.warning("Error Occurs While Getting the BASE_URLs")

# for sending urls to dataupload file
def getUrls(apiName):
        return urls["data"].get(apiName, None)

def checkUrl(url):
    try:
        l = url.index('.com')
        url = url[:l+4:] + "/enter-point"
        return url
    except Exception:
        traceback.print_exc()
        logging.warning("Error Occurs While handling the BASE_URLs")

count = 0
